"""Provide database clients."""

from .database import (
    AWS_ENV_VAR_NAME,
    AbstractDatabase,
    DatabaseException,
    DatabaseInitializationException,
    DatabaseReadException,
    DatabaseWriteException,
    create_db,
)

__all__ = [
    "AWS_ENV_VAR_NAME",
    "AbstractDatabase",
    "DatabaseException",
    "DatabaseInitializationException",
    "DatabaseReadException",
    "DatabaseWriteException",
    "create_db",
]



"""Provides abstract base classes for source ETL procedures.

Generally, users shouldn't ever have to work directly with the classes contained within.
"""

import logging
from abc import ABC, abstractmethod
from pathlib import Path

import click
from owlready2.rdflib_store import TripleLiteRDFlibGraph as RDFGraph
from wags_tails import CustomData, DataSource, DoData, MondoData, NcitData, OncoTreeData

from disease import ITEM_TYPES, SOURCES_FOR_MERGE
from disease.database import AbstractDatabase
from disease.etl.rules import Rules
from disease.schemas import Disease, SourceName

DATA_DISPATCH = {
    SourceName.NCIT: NcitData,
    SourceName.ONCOTREE: OncoTreeData,
    SourceName.MONDO: MondoData,
    SourceName.DO: DoData,
}

_logger = logging.getLogger(__name__)


class DiseaseNormalizerEtlError(Exception):
    """Basic ETL exception."""


class Base(ABC):
    """The ETL base class."""

    def __init__(
        self,
        database: AbstractDatabase,
        data_path: Path | None = None,
        silent: bool = True,
    ) -> None:
        """Extract from sources.

        :param database: database client
        :param data_path: location of data directory
        :param silent: if True, don't print ETL results to console
        """
        self._silent = silent
        self._src_name = SourceName(self.__class__.__name__)
        self._data_source: NcitData | OncoTreeData | MondoData | DoData | CustomData = (
            self._get_data_handler(data_path)
        )
        self._database = database
        self.store_ids = self.__class__.__name__ in SOURCES_FOR_MERGE
        self._rules = Rules(self._src_name)
        self._added_ids = []

    def _get_data_handler(self, data_path: Path | None = None) -> DataSource:
        """Construct data handler instance for source. Overwrite for edge-case sources.

        :param data_path: location of data storage
        :return: instance of wags_tails.DataSource to manage source file(s)
        """
        return DATA_DISPATCH[self._src_name](data_dir=data_path, silent=self._silent)

    def perform_etl(self, use_existing: bool = False) -> list:
        """Public-facing method to begin ETL procedures on given data.

        :param use_existing: if True, use local data instead of retrieving most recent
            version
        :return: List of concept IDs to be added to merge generation.
        """
        self._extract_data(use_existing)
        self._load_meta()
        if not self._silent:
            click.echo("Transforming and loading data to DB...")
        self._transform_data()
        self._database.complete_write_transaction()
        return self._added_ids

    def _extract_data(self, use_existing: bool = False) -> None:
        """Get source file from data directory.
        :param use_existing: if True, use local data regardless of whether it's up to
            date
        """
        self._data_file, self._version = self._data_source.get_latest(
            from_local=use_existing
        )

    @abstractmethod
    def _transform_data(self, *args, **kwargs) -> None:  # noqa: ANN002
        raise NotImplementedError

    @abstractmethod
    def _load_meta(self, *args, **kwargs) -> None:  # noqa: ANN002
        raise NotImplementedError

    def _load_disease(self, disease: dict) -> None:
        """Load individual disease record."""
        _ = Disease(**disease)
        concept_id = disease["concept_id"]

        for attr_type in ITEM_TYPES:
            if attr_type in disease:
                value = disease[attr_type]
                if value is not None and value != []:
                    if isinstance(value, str):
                        items = [value.lower()]
                    else:
                        disease[attr_type] = list(set(value))
                        items = {item.lower() for item in value}
                        if (attr_type == "aliases") and (len(items) > 20):
                            _logger.debug("%s has > 20 aliases.", concept_id)
                            del disease[attr_type]
                            continue

                else:
                    del disease[attr_type]

        for field in ("pediatric_disease", "oncologic_disease"):
            if field in disease and disease[field] is None:
                del disease[field]

        disease = self._rules.apply_rules_to_disease(disease)
        self._database.add_record(disease, self._src_name)
        self._added_ids.append(concept_id)


class OWLBase(Base):
    """Base class for sources that use OWL files."""

    def _get_subclasses(self, uri: str, graph: RDFGraph) -> set[str]:
        """Retrieve URIs for all terms that are subclasses of given URI.

        :param uri: URI for class
        :param graph: RDFLib graph of ontology default world
        :return: Set of URIs (strings) for all subclasses of `uri`
        """
        query = f"""
            SELECT ?c WHERE {{
                ?c rdfs:subClassOf* <{uri}>
            }}
            """
        return {item.c.toPython() for item in graph.query(query)}

    def _get_by_property_value(
        self, prop: str, value: str, graph: RDFGraph
    ) -> set[str]:
        """Get all classes with given value for a specific property.

        :param prop: property URI
        :param value: property value
        :param graph: RDFLib graph of ontology default world
        :return: Set of URIs (as strings) matching given property/value
        """
        query = f"""
            SELECT ?c WHERE {{
                ?c <{prop}>
                "{value}"
            }}
            """
        return {item.c.toPython() for item in graph.query(query)}



"""Provides query handler class, which receives and responses to user search queries."""

import datetime
import logging
import re

from botocore.exceptions import ClientError
from ga4gh.core.models import (
    Extension,
    MappableConcept,
    Relation,
)

from disease import NAMESPACE_LOOKUP, PREFIX_LOOKUP, SOURCES_LOWER_LOOKUP, __version__
from disease.database.database import AbstractDatabase
from disease.schemas import (
    Disease,
    MatchType,
    NormalizationService,
    RefType,
    SearchService,
    ServiceMeta,
    SourceName,
    get_coding_object,
    get_concept_mapping,
)

_logger = logging.getLogger(__name__)


class InvalidParameterException(Exception):  # noqa: N818
    """Exception for invalid parameter args provided by the user."""


class QueryHandler:
    """Class for normalizer management. Stores reference to database instance
    and normalizes query input.
    """

    def __init__(self, database: AbstractDatabase) -> None:
        """Initialize QueryHandler instance. Requires a created database object to
        initialize. The most straightforward way to do this is via the
        :py:meth:`~disease.database.database.create_db` method:

        >>> from disease.query import QueryHandler
        >>> from disease.database import create_db
        >>> q = QueryHandler(create_db())

        We'll generally call ``create_db`` without any arguments in code examples, for
        the sake of brevity. See the :py:meth:`~disease.database.database.create_db` API
        description for more details.

        :param database: storage backend to search against
        """
        self.db = database

    def _emit_warnings(self, query_str: str) -> dict | None:
        """Emit warnings if query contains non breaking space characters.
        :param str query_str: query string
        :return: dict keying warning type to warning description
        """
        warnings = None
        nbsp = re.search("\xa0|&nbsp;", query_str)
        if nbsp:
            warnings = {"nbsp": "Query contains non breaking space characters."}
            _logger.warning(
                "Query (%s) contains non breaking space characters.", query_str
            )
        return warnings

    def _add_record(
        self, response: dict[str, dict], item: dict, match_type: str
    ) -> tuple[dict, str]:
        """Add individual record to response object

        :param Dict[str, Dict] response: in-progress response object to return
            to client
        :param Dict item: Item retrieved from DynamoDB
        :param str match_type: type of query match
        :return: Tuple containing updated response object, and string
            containing name of the source of the match
        """
        disease = Disease(**item)
        src_name = item["src_name"]

        matches = response["source_matches"]
        if src_name not in matches:
            pass
        elif matches[src_name] is None:
            matches[src_name] = {
                "match_type": MatchType[match_type.upper()],
                "records": [disease],
                "source_meta_": self.db.get_source_metadata(src_name),
            }
        elif matches[src_name]["match_type"] == MatchType[match_type.upper()]:
            matches[src_name]["records"].append(disease)

        return response, src_name

    def _fetch_records(
        self, response: dict[str, dict], concept_ids: set[str], match_type: str
    ) -> tuple[dict, set]:
        """Return matched Disease records as a structured response for a given
        collection of concept IDs.

        :param Dict[str, Dict] response: in-progress response object to return
            to client.
        :param List[str] concept_ids: List of concept IDs to build from.
            Should be all lower-case.
        :param str match_type: record should be assigned this type of match.
        :return: response Dict with records filled in via provided concept
            IDs, and Set of source names of matched records
        """
        matched_sources = set()

        try:
            for concept_id in concept_ids:
                match = self.db.get_record_by_id(
                    concept_id.lower(), case_sensitive=False
                )
                if match is None:
                    _logger.error("Reference to %s failed.", concept_id)
                else:
                    (response, src) = self._add_record(response, match, match_type)
                    matched_sources.add(src)
        except ClientError as e:
            _logger.exception(e.response["Error"]["Message"])
        return response, matched_sources

    def _fill_no_matches(self, resp: dict[str, dict]) -> dict:
        """Fill all empty source_matches slots with NO_MATCH results.

        :param Dict[str, Dict] resp: incoming response object
        :return: response object with empty source slots filled with
                NO_MATCH results and corresponding source metadata
        """
        for src_name in resp["source_matches"]:
            if resp["source_matches"][src_name] is None:
                resp["source_matches"][src_name] = {
                    "match_type": MatchType.NO_MATCH,
                    "records": [],
                    "source_meta_": self.db.get_source_metadata(src_name),
                }
        return resp

    def _check_concept_id(
        self, query: str, resp: dict, sources: set[str]
    ) -> tuple[dict, set]:
        """Check query for concept ID match. Should only find 0 or 1 matches.

        :param str query: search string
        :param Dict resp: in-progress response object to return to client
        :param Set[str] sources: remaining unmatched sources
        :return: Tuple with updated resp object and updated set of unmatched
            sources
        """
        concept_id_items = []
        if [p for p in PREFIX_LOOKUP if query.startswith(p)]:
            record = self.db.get_record_by_id(query, False)
            if record:
                concept_id_items.append(record)
        for prefix in [p for p in NAMESPACE_LOOKUP if query.startswith(p)]:
            concept_id = f"{NAMESPACE_LOOKUP[prefix]}:{query}"
            id_lookup = self.db.get_record_by_id(concept_id, False)
            if id_lookup:
                concept_id_items.append(id_lookup)
        for item in concept_id_items:
            (resp, src_name) = self._add_record(resp, item, MatchType.CONCEPT_ID.name)
            sources = sources - {src_name}
        return resp, sources

    def _check_match_type(
        self, query: str, resp: dict, sources: set[str], match_type: RefType
    ) -> tuple[dict, set]:
        """Check query for selected match type.

        :param str query: search string
        :param Dict resp: in-progress response object to return to client
        :param Set[str] sources: remaining unmatched sources
        :param RefType match_type: Match type to check for. Should be one of
            {'label', 'alias', 'xref', 'associated_with'}
        :return: Tuple with updated resp object and updated set of unmatched
                 sources
        """
        matching_ids = self.db.get_refs_by_type(query, match_type)
        if matching_ids:
            (resp, matched_srcs) = self._fetch_records(
                resp, set(matching_ids), match_type
            )
            sources = sources - matched_srcs
        return resp, sources

    def _get_search_response(self, query: str, sources: set[str]) -> dict:
        """Return response as dict where key is source name and value is a list of
        records.

        :param str query: string to match against
        :param Set[str] sources: sources to match from
        :return: completed response object to return to client
        """
        response = {
            "query": query,
            "warnings": self._emit_warnings(query),
            "source_matches": dict.fromkeys(sources),
        }
        if query == "":
            return self._fill_no_matches(response)
        query = query.lower()

        # check if concept ID match
        (response, sources) = self._check_concept_id(query, response, sources)
        if len(sources) == 0:
            return response

        for match_type in RefType:
            (response, sources) = self._check_match_type(
                query, response, sources, match_type
            )
            if len(sources) == 0:
                return response

        # remaining sources get no match
        return self._fill_no_matches(response)

    def search(self, query_str: str, incl: str = "", excl: str = "") -> SearchService:
        """Return highest match for each source.

        >>> from disease.query import QueryHandler
        >>> from disease.database import create_db
        >>> q = QueryHandler(create_db())
        >>> result = q.search("NSCLC")
        >>> result.source_matches["Mondo"].records[0].concept_id
        'mondo:0005233'

        :param str query_str: query, a string, to search for
        :param str incl: str containing comma-separated names of sources to
            use. Will exclude all other sources. Case-insensitive.
        :param str excl: str containing comma-separated names of source to
            exclude. Will include all other source. Case-insensitive.
        :return: dict containing all matches found in sources.
        :rtype: dict
        :raises InvalidParameterException: if both incl and excl args are
            provided, or if invalid source names are given.
        """
        sources = {
            k: v
            for k, v in SOURCES_LOWER_LOOKUP.items()
            if self.db.get_source_metadata(v)
        }

        if not incl and not excl:
            query_sources = set(sources.values())
        elif incl and excl:
            detail = "Cannot request both source inclusions and exclusions."
            raise InvalidParameterException(detail)
        elif incl:
            req_sources = [n.strip() for n in incl.split(",")]
            invalid_sources = []
            query_sources = set()
            for source in req_sources:
                if source.lower() in sources:
                    query_sources.add(sources[source.lower()])
                else:
                    invalid_sources.append(source)
            if invalid_sources:
                detail = f"Invalid source name(s): {invalid_sources}"
                raise InvalidParameterException(detail)
        else:
            req_exclusions = [n.strip() for n in excl.lower().split(",")]
            req_excl_dict = {r.lower(): r for r in req_exclusions}
            invalid_sources = []
            query_sources = set()
            for req_l, req in req_excl_dict.items():
                if req_l not in sources:
                    invalid_sources.append(req)
            for src_l, src in sources.items():
                if src_l not in req_excl_dict:
                    query_sources.add(src)
            if invalid_sources:
                detail = f"Invalid source name(s): {invalid_sources}"
                raise InvalidParameterException(detail)

        query_str = query_str.strip()

        response = self._get_search_response(query_str, query_sources)

        response["service_meta_"] = ServiceMeta(
            version=__version__,
            response_datetime=datetime.datetime.now(tz=datetime.UTC),
        ).model_dump()
        return SearchService(**response)

    def _add_merged_meta(self, response: dict) -> dict:
        """Add source metadata to response object.

        :param Dict response: in-progress response object
        :return: completed response object.
        """

        def _get_src_prefix(identifier: str) -> str | None:
            """Get source name prefix for an identifier

            :param identifier: Identifier for a given system
            :return: Source name prefix, if it's a source we ingest (in `PREFIX_LOOKUP`)
            """
            ns = re.split(r"[:_]", identifier, maxsplit=1)[0].lower()
            return PREFIX_LOOKUP.get(ns)

        sources_meta = {}
        disease = response["disease"]

        sources = [_get_src_prefix(disease.primaryCoding.id)]
        for m in disease.mappings or []:
            m_src = _get_src_prefix(m.coding.id)
            if m_src:
                sources.append(m_src)

        for src in sources:
            if src not in sources_meta:
                sources_meta[src] = self.db.get_source_metadata(src)
        response["source_meta_"] = sources_meta
        return response

    def _add_disease(
        self, response: dict, record: dict, match_type: MatchType
    ) -> NormalizationService:
        """Format received DB record as core Disease object and update response object.

        :param response: in-progress response object
        :param record: record as stored in DB
        :param match_type: type of match achieved
        :return: completed normalized response object ready to return to user
        """
        disease_obj = MappableConcept(
            id=f"normalize.disease.{record['concept_id']}",
            primaryCoding=get_coding_object(record["concept_id"]),
            conceptType="Disease",
            name=record["label"],
            extensions=[],
        )

        xrefs = record.get("xrefs", [])
        mappings = [
            get_concept_mapping(xref_id, relation=Relation.EXACT_MATCH)
            for xref_id in xrefs
        ]

        associated_with = record.get("associated_with", [])
        mappings.extend(
            get_concept_mapping(associated_with_id, relation=Relation.RELATED_MATCH)
            for associated_with_id in associated_with
        )

        disease_obj.mappings = mappings or None

        for field in ("pediatric_disease", "oncologic_disease", "aliases"):
            value = record.get(field)
            if value is not None:
                disease_obj.extensions.append(Extension(name=field, value=value))

        response["match_type"] = match_type
        response["disease"] = disease_obj
        response = self._add_merged_meta(response)
        return NormalizationService(**response)

    def _record_order(self, record: dict) -> tuple[int, str]:
        """Construct priority order for matching. Only called by sort().

        :param Dict record: individual record item in iterable to sort
        :return: tuple with rank value and concept ID
        """
        src = record["src_name"]
        if src == SourceName.NCIT.value:
            source_rank = 1
        elif src == SourceName.MONDO.value:
            source_rank = 2
        elif src == SourceName.ONCOTREE.value:
            source_rank = 3
        elif src == SourceName.OMIM.value:
            source_rank = 4
        elif src == SourceName.DO.value:
            source_rank = 5
        else:
            _logger.warning("query.record_order: Invalid source name for %s", record)
            source_rank = 4
        return source_rank, record["concept_id"]

    def _handle_failed_merge_ref(
        self, record: dict, response: dict, query: str
    ) -> NormalizationService:
        """Log + fill out response for a failed merge reference lookup.

        :param Dict record: record containing failed merge_ref
        :param Dict response: in-progress response object
        :param str query: original query value
        :return: Normalized response with no match
        """
        _logger.error(
            "Merge ref lookup failed for ref %s in record %s from query %s",
            record["merge_ref"],
            record["concept_id"],
            query,
        )
        response["match_type"] = MatchType.NO_MATCH
        return NormalizationService(**response)

    def normalize(self, query: str) -> NormalizationService:
        """Return normalized concept for query.

        Use to retrieve normalized disease concept records:

        >>> from disease.query import QueryHandler
        >>> from disease.database import create_db
        >>> q = QueryHandler(create_db())
        >>> result = q.normalize("NSCLC")
        >>> result.disease.primaryCoding.id
        'ncit:C2926'

        :param query: String to find normalized concept for
        :return: Normalized disease concept
        """
        # prepare basic response
        response = {
            "match_type": MatchType.NO_MATCH,
            "query": query,
            "warnings": self._emit_warnings(query),
            "service_meta_": ServiceMeta(
                version=__version__,
                response_datetime=datetime.datetime.now(tz=datetime.UTC),
            ),
        }
        if query == "":
            return NormalizationService(**response)
        query_str = query.lower().strip()

        # check merged concept ID match
        record = self.db.get_record_by_id(query_str, case_sensitive=False, merge=True)
        if record:
            return self._add_disease(response, record, MatchType.CONCEPT_ID)

        non_merged_match = None

        # check for concept ID match against record that has a ref to a merged record
        record = self.db.get_record_by_id(query_str, case_sensitive=False)
        if record:
            if "merge_ref" in record:
                merge = self.db.get_record_by_id(
                    record["merge_ref"], case_sensitive=False, merge=True
                )
                if merge is None:
                    return self._handle_failed_merge_ref(record, response, query_str)
                return self._add_disease(response, merge, MatchType.CONCEPT_ID)
            non_merged_match = (record, "concept_id")

        # check other match types
        for match_type in RefType:
            # get matches list for match tier
            matching_refs = self.db.get_refs_by_type(query_str, match_type)
            matching_records = [
                self.db.get_record_by_id(ref, case_sensitive=False)
                for ref in matching_refs
            ]
            matching_records.sort(key=self._record_order)

            # attempt merge ref resolution until successful
            for match in matching_records:
                record = self.db.get_record_by_id(match["concept_id"], merge=False)
                if record:
                    if "merge_ref" in record:
                        merge = self.db.get_record_by_id(
                            record["merge_ref"], case_sensitive=False, merge=True
                        )
                        if merge is None:
                            return self._handle_failed_merge_ref(
                                record, response, query_str
                            )
                        return self._add_disease(
                            response, merge, MatchType[match_type.upper()]
                        )
                    if not non_merged_match:
                        non_merged_match = (record, match_type)

        # if no successful match, try best available non-merged match
        if non_merged_match:
            match_type = MatchType[non_merged_match[1].upper()]
            return self._add_disease(response, non_merged_match[0], match_type)

        return NormalizationService(**response)



"""Get Mondo Disease Ontology data."""

import logging
import re
from collections import defaultdict
from typing import ClassVar

import fastobo
from tqdm import tqdm

from disease.etl.base import Base
from disease.schemas import DataLicenseAttributes, NamespacePrefix, SourceMeta

_logger = logging.getLogger(__name__)


class Mondo(Base):
    """Gather and load data from Mondo."""

    def _load_meta(self) -> None:
        """Load metadata"""
        metadata = SourceMeta(
            data_license="CC BY 4.0",
            data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",
            version=self._version,
            data_url="https://mondo.monarchinitiative.org/pages/download/",
            rdp_url="http://reusabledata.org/monarch.html",
            data_license_attributes=DataLicenseAttributes(
                non_commercial=False, share_alike=False, attribution=True
            ),
        )
        self._database.add_source_metadata(self._src_name, metadata)

    def _construct_dependency_set(self, dag: defaultdict, parent: str) -> set[str]:
        """Recursively get all children concepts for a term

        :param dag: dictionary where keys are ontology terms and values are lists of
            terms with ``is_a`` relationships to the parent
        :param parent: term to fetch children for
        :return: Set of children concepts
        """
        children = {parent}
        for child in dag[parent]:
            children |= self._construct_dependency_set(dag, child)
        return children

    _identifiers_url_pattern = r"http://identifiers.org/(.*)/(.*)"
    _lui_patterns: ClassVar = [
        # (NamespacePrefix.OMIMPS, r"https://omim.org/phenotypicSeries/(.*)"),
        (NamespacePrefix.OMIM, r"https://omim.org/entry/(.*)"),
        (NamespacePrefix.UMLS, r"http://linkedlifedata.com/resource/umls/id/(.*)"),
        (NamespacePrefix.ICD10CM, r"http://purl.bioontology.org/ontology/ICD10CM/(.*)"),
        (NamespacePrefix.ICD10, r"https://icd.who.int/browse10/2019/en#/(.*)"),
    ]

    def _get_xref_from_url(self, url: str) -> tuple[NamespacePrefix, str] | None:
        """Extract prefix and LUI from URL reference.

        :param url: url string given as URL xref property
        :return: prefix enum instance and LUI
        """
        if url.startswith("http://identifiers.org"):
            match = re.match(self._identifiers_url_pattern, url)
            if not match or not match.groups():
                msg = f"Couldn't parse identifiers.org URL: {url}"
                raise ValueError(msg)
            if match.groups()[0] == "snomedct":
                return None
            prefix = NamespacePrefix[match.groups()[0].upper()]
            return (prefix, match.groups()[1])
        for prefix, pattern in self._lui_patterns:
            match = re.match(pattern, url)
            if match and match.groups():
                return (prefix, match.groups()[0])
        # didn't match any patterns
        _logger.warning("Unrecognized URL for xref: %s", url)
        return None

    @staticmethod
    def _get_xref_from_xref_clause(
        clause: fastobo.term.XrefClause,
    ) -> tuple[NamespacePrefix, str] | None:
        """Get dbXref from xref clause.

        In the Mondo OBO distribution, some xrefs only show up in explicit xref clauses.
        This method processes them.

        :param clause: xref clause from term frame
        :return: prefix and local ID if available
        """
        raw_prefix = clause.xref.id.prefix
        if raw_prefix not in ("ONCOTREE", "EFO"):
            return None
        try:
            prefix = NamespacePrefix[clause.xref.id.prefix.upper()]
        except KeyError:
            _logger.warning("Unable to parse namespace prefix for %s", clause.xref)
            return None
        local_id = clause.xref.id.local
        return prefix, local_id

    def _get_xref_from_pv_clause(
        self, clause: fastobo.term.PropertyValueClause
    ) -> tuple[NamespacePrefix, str] | None:
        """Get dbXref from property value clause.

        These are a bit more semantically rich than the Mondo xref clauses, so for now,
        we prefer to fetch most references from here.

        :param clause: property value clause
        :return: prefix and local ID if available
        """
        property_value = clause.property_value
        # of all property values, all we care about are those which are skos:exactMatch,
        # so we can ignore other relations and any non-resource property values.
        # previously there may have been usage of mondo:equivalentTo but that seems to
        # be irrelevant for property values
        if not isinstance(
            property_value, fastobo.pv.ResourcePropertyValue
        ) or property_value.relation != fastobo.id.PrefixedIdent("skos", "exactMatch"):
            return None
        if isinstance(property_value.value, fastobo.id.Url):
            xref_result = self._get_xref_from_url(str(property_value.value))
            if xref_result is None:
                return None
            prefix, local_id = xref_result
        elif isinstance(property_value.value, fastobo.id.PrefixedIdent):
            prefix = NamespacePrefix[property_value.value.prefix.upper()]
            local_id = property_value.value.local
        else:
            _logger.warning(
                "Unrecognized property value type: %s", type(property_value.value)
            )
            return None
        return prefix, local_id

    def _process_term_frame(self, frame: fastobo.term.TermFrame) -> dict:
        """Extract disease params from an OBO term frame.

        :param frame: individual frame from OBO file
        :return: disease params as a dictionary
        """
        params = {
            "concept_id": str(frame.id).lower(),
            "aliases": [],
            "xrefs": [],
            "associated_with": [],
        }

        for clause in frame:
            tag = clause.raw_tag()
            if tag == "name":
                params["label"] = clause.raw_value()
            elif tag == "synonym":
                params["aliases"].append(clause.synonym.desc)
            elif tag in ("xref", "property_value"):
                if tag == "xref":
                    xref = self._get_xref_from_xref_clause(clause)
                else:
                    xref = self._get_xref_from_pv_clause(clause)
                if not xref:
                    continue
                prefix, local_id = xref

                curie = f"{prefix.value}:{local_id}"
                if prefix in (
                    NamespacePrefix.OMIM,
                    NamespacePrefix.NCIT,
                    NamespacePrefix.DO,
                    NamespacePrefix.ONCOTREE,
                ):
                    params["xrefs"].append(curie)
                else:
                    params["associated_with"].append(curie)
        return params

    def _transform_data(self) -> None:
        """Get data from file and send disease records to database."""
        reader = fastobo.iter(str(self._data_file.absolute()))
        dag = defaultdict(list)
        for item in reader:
            item_id = str(item.id)
            for clause in item:
                if clause.raw_tag() == "is_a":
                    dag[clause.raw_value()].append(item_id)

        disease_root = "MONDO:0000001"
        diseases = self._construct_dependency_set(dag, disease_root)
        peds_neoplasm_root = "MONDO:0006517"
        pediatric_diseases = self._construct_dependency_set(dag, peds_neoplasm_root)
        cancer_root = "MONDO:0045024"
        cancers = self._construct_dependency_set(dag, cancer_root)

        reader = fastobo.iter(str(self._data_file.absolute()))
        for item in tqdm(reader, ncols=80, disable=self._silent):
            concept_id = str(item.id).lower()
            if concept_id.upper() not in diseases:
                continue

            params = self._process_term_frame(item)

            if concept_id.upper() in pediatric_diseases:
                params["pediatric_disease"] = True

            if concept_id.upper() in cancers:
                params["oncologic_disease"] = True

            self._load_disease(params)



"""Get OMIM disease data."""

from pathlib import Path

from tqdm import tqdm
from wags_tails import CustomData, DataSource

from disease.etl.base import Base
from disease.schemas import NamespacePrefix, SourceMeta


class OMIM(Base):
    """Gather and load data from OMIM."""

    def _raise_access_error(self) -> None:
        """Raise improper data access error and describe proper data access."""
        msg = f"Could not locate OMIM data. Per README, OMIM source files must be manually placed in {self._data_source.data_dir.absolute().as_uri()}"
        raise FileNotFoundError(msg)

    def _get_data_handler(self, data_path: Path | None = None) -> DataSource:
        """Construct data handler instance for source. Overwrites base class method
        to use custom data handler instead.

        :param data_path: location of data storage
        :return: instance of wags_tails.DataSource to manage source file(s)
        """
        return CustomData(
            "omim",
            "tsv",
            latest_version_cb=lambda: "",
            download_cb=lambda: self._raise_access_error(),
            data_dir=data_path,
        )

    def _extract_data(self, use_existing: bool = False) -> None:  # noqa: ARG002
        """Get source file from data directory.

        :param use_existing: if True, use local data regardless of whether it's up to
            date. OMIM data must be manually provided, so this argument is ignored.
        """
        self._data_file, self._version = self._data_source.get_latest(from_local=True)

    def _load_meta(self) -> None:
        """Load source metadata."""
        metadata = SourceMeta(
            data_license="custom",
            data_license_url="https://omim.org/help/agreement",
            version=self._data_file.stem.split("_", 1)[1],
            data_url="https://www.omim.org/downloads",
            rdp_url="http://reusabledata.org/omim.html",
            data_license_attributes={
                "non_commercial": False,
                "share_alike": True,
                "attribution": True,
            },
        )
        self._database.add_source_metadata(self._src_name, metadata)

    def _transform_data(self) -> None:
        """Modulate data and prepare for loading."""
        with self._data_file.open() as f:
            rows = f.readlines()
        rows = [r.rstrip() for r in rows if not r.startswith("#")]
        rows = [r.split("\t") for r in rows]
        rows = [r for r in rows if r[0] not in ("Asterisk", "Caret", "Plus")]
        for row in tqdm(rows, ncols=80, disable=self._silent):
            disease = {
                "concept_id": f"{NamespacePrefix.OMIM.value}:{row[1]}",
            }
            aliases = set()

            label_item = row[2]
            if ";" in label_item:
                label_split = label_item.split(";")
                disease["label"] = label_split[0]
                aliases.add(label_split[1])
            else:
                disease["label"] = row[2]

            if len(row) > 3:
                aliases |= {t for t in row[3].split(";") if t}
            if len(row) > 4:
                aliases |= {t for t in row[4].split(";") if t}
            aliases = {alias.removesuffix(", INCLUDED") for alias in aliases}
            disease["aliases"] = [a.lstrip() for a in aliases]

            self._load_disease(disease)



"""Pytest test config tools."""

import logging
import os
from collections.abc import Generator
from pathlib import Path
from typing import Any

import pytest

from disease.config import get_config
from disease.database.database import AWS_ENV_VAR_NAME, AbstractDatabase, create_db
from disease.etl.base import Base
from disease.query import QueryHandler
from disease.schemas import (
    Disease,
    MatchType,
    RecordType,
    RefType,
    SourceMeta,
    SourceName,
    SourceSearchMatches,
)

_logger = logging.getLogger(__name__)


def pytest_collection_modifyitems(items):
    """Modify test items in place to ensure test modules run in a given order.

    When creating new test modules, be sure to add them here.
    """
    module_order = [
        "test_schemas",
        "test_mondo",
        "test_do",
        "test_ncit",
        "test_omim",
        "test_oncotree",
        "test_merge",
        "test_database",
        "test_query",
        "test_api",
        "test_emit_warnings",
        "test_utils",
    ]
    # remember to add new test modules to the order constant:
    assert len(module_order) == len(list(Path(__file__).parent.rglob("test_*.py")))
    items.sort(key=lambda i: module_order.index(i.module.__name__))


def pytest_addoption(parser):
    """Add custom commands to pytest invocation.
    See https://docs.pytest.org/en/7.1.x/reference/reference.html#parser
    """
    parser.addoption(
        "--verbose-logs",
        action="store_true",
        default=False,
        help="show noisy module logs",
    )


def pytest_configure(config):
    """Configure pytest setup."""
    if not config.getoption("--verbose-logs"):
        logging.getLogger("botocore").setLevel(logging.ERROR)
        logging.getLogger("boto3").setLevel(logging.ERROR)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)


@pytest.fixture(scope="session")
def test_data_dir() -> Path:
    """Provide Path instance pointing to test data directory"""
    return Path(__file__).parent / "data"


def pytest_sessionstart():
    """Wipe DB before testing if in test environment."""
    if get_config().test:
        if os.environ.get(AWS_ENV_VAR_NAME):
            msg = f"Cannot have both DISEASE_NORM_TEST and {AWS_ENV_VAR_NAME} set."
            raise AssertionError(msg)
        db = create_db()
        db.drop_db()
        db.initialize_db()


@pytest.fixture(scope="session")
def is_test_env():
    """If true, currently in test environment (i.e. okay to overwrite DB). Downstream
    users should also make sure to check if in a production environment.

    Provided here to be accessible directly within test modules.
    """
    return get_config().test


@pytest.fixture(scope="module")
def database():
    """Provide a database instance to be used by tests."""
    db = create_db()
    yield db
    db.close_connection()


@pytest.fixture(scope="module")
def test_source(database: AbstractDatabase, is_test_env: bool, test_data_dir: Path):
    """Provide query endpoint for testing sources. If DISEASE_NORM_TEST is set, will try to
    load DB from test data.

    :param database: test database instance
    :param is_test_env: if true, load from test data
    :return: factory function that takes an ETL class instance and returns a query
    endpoint.
    """

    def test_source_factory(EtlClass: Base):  # noqa: N803
        if is_test_env:
            _logger.debug("Reloading DB with data from %s", test_data_dir)
            test_class = EtlClass(database, test_data_dir / EtlClass.__name__.lower())
            test_class.perform_etl(use_existing=True)

        class QueryGetter:
            def __init__(self):
                self.query_handler = QueryHandler(database)
                self._src_name = EtlClass.__name__

            def search(self, query_str: str):
                resp = self.query_handler.search(query_str, incl=self._src_name)
                return resp.source_matches[self._src_name]

        return QueryGetter()

    return test_source_factory


def _compare_records(actual: Disease, fixt: Disease):
    """Check that identity records are identical."""
    assert actual.concept_id == fixt.concept_id
    assert actual.label == fixt.label

    assert (actual.aliases is None) == (fixt.aliases is None)
    if (actual.aliases is not None) and (fixt.aliases is not None):
        assert set(actual.aliases) == set(fixt.aliases)

    assert (actual.xrefs is None) == (fixt.xrefs is None)
    if (actual.xrefs is not None) and (fixt.xrefs is not None):
        assert set(actual.xrefs) == set(fixt.xrefs)

    assert (actual.associated_with is None) == (fixt.associated_with is None)
    if (actual.associated_with is not None) and (fixt.associated_with is not None):
        assert set(actual.associated_with) == set(fixt.associated_with)

    assert (actual.pediatric_disease is None) == (fixt.pediatric_disease is None)
    if (actual.pediatric_disease is not None) and (fixt.pediatric_disease is not None):
        assert actual.pediatric_disease == fixt.pediatric_disease

    assert (actual.oncologic_disease is None) == (fixt.oncologic_disease is None)
    if (actual.oncologic_disease is not None) and (fixt.oncologic_disease is not None):
        assert actual.oncologic_disease == fixt.oncologic_disease


@pytest.fixture(scope="session")
def compare_records():
    """Provide record comparison function"""
    return _compare_records


def _compare_response(
    response: SourceSearchMatches,
    match_type: MatchType,
    fixture: Disease | None = None,
    fixture_list: list[Disease] | None = None,
    num_records: int = 0,
):
    """Check that test response is correct. Only 1 of {fixture, fixture_list}
    should be passed as arguments. num_records should only be passed with fixture_list.

    :param Dict response: response object returned by QueryHandler
    :param MatchType match_type: expected match type
    :param Disease fixture: single Disease object to match response against
    :param List[Disease] fixture_list: multiple Disease objects to match response
        against
    :param int num_records: expected number of records in response. If not given, tests
        for number of fixture Diseases given (ie, 1 for single fixture and length of
        fixture_list otherwise)
    """
    if fixture and fixture_list:
        pytest.fail("Args provided for both `fixture` and `fixture_list`")
    if not fixture and not fixture_list:
        pytest.fail("Must pass 1 of {fixture, fixture_list}")
    if fixture and num_records:
        pytest.fail("`num_records` should only be given with `fixture_list`.")

    assert response.match_type == match_type
    if fixture:
        assert len(response.records) == 1
        _compare_records(response.records[0], fixture)
    elif fixture_list:
        if not num_records:
            assert len(response.records) == len(fixture_list)
        else:
            assert len(response.records) == num_records
        for fixt in fixture_list:
            for record in response.records:
                if fixt.concept_id == record.concept_id:
                    _compare_records(record, fixt)
                    break
            raise AssertionError  # test fixture not found in response


@pytest.fixture(scope="session")
def compare_response():
    """Provide response comparison function"""
    return _compare_response


@pytest.fixture(scope="session")
def null_database_class():
    """Quote-unquote 'in-memory database' used like a mock for testing.

    Parameters for specific methods enabled as needed. See `tests/unit/test_utils.py`
    for example usage.
    """

    class _Database(AbstractDatabase):
        def __init__(self, db_url: str | None = None, **db_args) -> None:  # noqa: ARG002
            self._get_all_records_values = db_args.get("get_all_records", {})

        def list_tables(self) -> list[str]:
            raise NotImplementedError

        def drop_db(self) -> None:
            raise NotImplementedError

        def check_schema_initialized(self) -> bool:
            raise NotImplementedError

        def check_tables_populated(self) -> bool:
            raise NotImplementedError

        def initialize_db(self) -> None:
            raise NotImplementedError

        def get_source_metadata(self, src_name: str | SourceName) -> SourceMeta | None:
            raise NotImplementedError

        def get_record_by_id(
            self, concept_id: str, case_sensitive: bool = True, merge: bool = False
        ) -> dict | None:
            raise NotImplementedError

        def get_refs_by_type(self, search_term: str, ref_type: RefType) -> list[str]:
            raise NotImplementedError

        def get_all_concept_ids(self, source: SourceName | None = None) -> set[str]:
            raise NotImplementedError

        def get_all_records(
            self, record_type: RecordType
        ) -> Generator[dict, None, None]:
            yield from self._get_all_records_values[record_type]

        def add_source_metadata(self, src_name: SourceName, meta: SourceMeta) -> None:
            raise NotImplementedError

        def add_record(self, record: dict, src_name: SourceName) -> None:
            raise NotImplementedError

        def add_merged_record(self, record: dict) -> None:
            raise NotImplementedError

        def update_merge_ref(self, concept_id: str, merge_ref: Any) -> None:  # noqa: ANN401
            raise NotImplementedError

        def delete_normalized_concepts(self) -> None:
            raise NotImplementedError

        def delete_source(self, src_name: SourceName) -> None:
            raise NotImplementedError

        def complete_write_transaction(self) -> None:
            raise NotImplementedError

        def close_connection(self) -> None:
            raise NotImplementedError

        def load_from_remote(self, url: str | None = None) -> None:
            raise NotImplementedError

        def export_db(self, export_location: Path) -> None:
            raise NotImplementedError

    return _Database



"""Construct test data for Mondo source."""

from collections import defaultdict
from pathlib import Path

import fastobo

from disease.database import create_db
from disease.etl import Mondo

mondo = Mondo(create_db())
mondo._extract_data()
infile = str(mondo._data_file.absolute())

scripts_dir = Path(__file__).resolve().parent
test_data_dir = scripts_dir.parent / "data" / "mondo"

reader = fastobo.iter(infile)
dag = defaultdict(list)
for frame in reader:
    item_id = str(frame.id)
    for clause in frame:
        if clause.raw_tag() == "is_a":
            dag[item_id].append(clause.raw_value())


def construct_inheritance_set(dag: defaultdict, child: str) -> set[str]:
    """Get IDs for concepts that a child concept inherits from

    :param dag: dictionary keying IDs to parent concepts
    :param child: concept to fetch parents of
    :return: set of parent concepts
    """
    parents = {child}
    for parent in dag[child]:
        parents |= construct_inheritance_set(dag, parent)
    return parents


relevant_terms = set()
for term in (
    "MONDO:0005072",
    "MONDO:0002083",
    "MONDO:0003587",
    "MONDO:0004099",
    "MONDO:0005233",
    "MONDO:0010648",
    "MONDO:0009539",
    "MONDO:0013108",
    "MONDO:0013082",
):
    relevant_terms |= construct_inheritance_set(dag, term)

outfile = test_data_dir / mondo._data_file.name
outfile.parent.mkdir(exist_ok=True)
with outfile.open("w") as f:
    doc = fastobo.load(infile)
    f.write(str(doc.header))
    f.write("\n")

    reader = fastobo.iter(infile)
    for frame in reader:
        if isinstance(frame, fastobo.term.TermFrame):
            if str(frame.id) in relevant_terms:
                f.write(str(frame))
                f.write("\n")
        else:
            f.write(str(frame))
            f.write("\n")



"""Provides a CLI util to make updates to normalizer database."""

import datetime
import json
import logging
import os
from pathlib import Path

import click

from disease import __version__
from disease.config import get_config
from disease.database.database import DatabaseException, create_db
from disease.schemas import RecordType, SourceName
from disease.utils import get_term_mappings, initialize_logs

_logger = logging.getLogger(__name__)


URL_DESCRIPTION = 'URL endpoint for the application database. Can either be a URL to a local DynamoDB server (e.g. "http://localhost:8001") or a libpq-compliant PostgreSQL connection description (e.g. "postgresql://postgres:password@localhost:5432/disease_normalizer").'
SILENT_MODE_DESCRIPTION = "Suppress output to console."


def _initialize_app() -> None:
    if get_config().debug:
        initialize_logs(logging.DEBUG)
    else:
        initialize_logs()


@click.group()
@click.version_option(__version__)
def cli() -> None:
    """Manage Disease Normalizer data."""


@cli.command()
@click.option("--db_url", help=URL_DESCRIPTION)
@click.option("--silent", is_flag=True, default=False, help=SILENT_MODE_DESCRIPTION)
def check_db(db_url: str, silent: bool) -> None:
    """Perform basic checks on DB health and population. Exits with status code 1
    if DB schema is uninitialized or if critical tables appear to be empty.

        $ disease-normalizer check-db
        $ echo $?
        1  # indicates failure

    This command is equivalent to the combination of the database classes'
    ``check_schema_initialized()`` and ``check_tables_populated()`` methods:

    >>> from disease.database import create_db
    >>> db = create_db()
    >>> db.check_schema_initialized() and db.check_tables_populated()
    True  # DB passes checks

    \f
    :param db_url: URL to normalizer database
    :param silent: if true, suppress console output
    """  # noqa: D301
    _initialize_app()
    db = create_db(db_url, False)
    if not db.check_schema_initialized():
        if not silent:
            click.echo("Health check failed: DB schema uninitialized.")
        click.get_current_context().exit(1)

    if not db.check_tables_populated():
        if not silent:
            click.echo("Health check failed: DB is incompletely populated.")
        click.get_current_context().exit(1)

    msg = "DB health check successful: tables appear complete."
    if not silent:
        click.echo(msg)
    _logger.info(msg)


@cli.command()
@click.option("--data_url", help="URL to data dump")
@click.option("--db_url", help=URL_DESCRIPTION)
@click.option("--silent", is_flag=True, default=False, help=SILENT_MODE_DESCRIPTION)
def update_from_remote(data_url: str | None, db_url: str, silent: bool) -> None:
    """Update data from remotely-hosted DB dump. By default, fetches from latest
    available dump on VICC S3 bucket; specific URLs can be provided instead by
    command line option or DISEASE_NORM_REMOTE_DB_URL environment variable.

    \f
    :param data_url: user-specified location to pull DB dump from
    :param db_url: URL to normalizer database
    :param silent: if true, suppress console output
    """  # noqa: D301
    _initialize_app()
    if not click.confirm("Are you sure you want to overwrite existing data?"):
        click.get_current_context().exit()
    if not data_url:
        data_url = os.environ.get("DISEASE_NORM_REMOTE_DB_URL")
    db = create_db(db_url, False)
    try:
        db.load_from_remote(data_url)
    except NotImplementedError:
        if not silent:
            click.echo(
                f"Error: Fetching remote data dump not supported for {db.__class__.__name__}"
            )
        click.get_current_context().exit(1)
    except DatabaseException as e:
        if not silent:
            click.echo(f"Encountered exception during update: {e!s}")
        _logger.exception(
            "Encountered exception. `data_url`=%s, `db_url`=%s", data_url, db_url
        )
        click.get_current_context().exit(1)
    _logger.info("Successfully loaded data from remote snapshot.")


@cli.command()
@click.option(
    "--output_directory",
    "-o",
    help="Output location to write to",
    type=click.Path(exists=True, path_type=Path),
)
@click.option("--db_url", help=URL_DESCRIPTION)
@click.option("--silent", is_flag=True, default=False, help=SILENT_MODE_DESCRIPTION)
def dump_database(output_directory: Path, db_url: str, silent: bool) -> None:
    """Dump data from database into file.

    \f
    :param output_directory: path to existing directory
    :param db_url: URL to normalizer database
    :param silent: if True, suppress output to console
    """  # noqa: D301
    _initialize_app()
    if not output_directory:
        output_directory = Path()

    db = create_db(db_url, False)
    try:
        db.export_db(output_directory)
    except NotImplementedError:
        msg = f"Error: Dumping data to file not supported for {db.__class__.__name__}"
        if not silent:
            click.echo(msg)
        _logger.exception(msg)
        click.get_current_context().exit(1)
    except DatabaseException as e:
        if not silent:
            click.echo(f"Encountered exception during update: {e!s}")
        _logger.exception(
            "Encountered exception. `data_url`=%s, `db_url`=%s",
            output_directory,
            db_url,
        )
        click.get_current_context().exit(1)
    _logger.info("Database dump successful.")


@cli.command()
@click.argument("sources", nargs=-1)
@click.option("--all", "all_", is_flag=True, help="Update records for all sources.")
@click.option("--normalize", is_flag=True, help="Create normalized records.")
@click.option("--db_url", help=URL_DESCRIPTION)
@click.option("--aws_instance", is_flag=True, help="Use cloud DynamodDB instance.")
@click.option(
    "--use_existing",
    is_flag=True,
    default=False,
    help="Use most recent locally-available source data instead of fetching latest version",
)
@click.option("--silent", is_flag=True, default=False, help=SILENT_MODE_DESCRIPTION)
def update(
    sources: tuple[str, ...],
    aws_instance: bool,
    db_url: str,
    all_: bool,
    normalize: bool,
    use_existing: bool,
    silent: bool,
) -> None:
    """Update provided normalizer SOURCES in the disease database.

    Valid SOURCES are "DO", "MONDO", "NCIt", "OMIM", and "OncoTree" (case is irrelevant).

    SOURCES are optional, but if not provided, either --all or --normalize must be used.

    For example, the following command will update DO and MONDO source records:

        $ disease-normalizer update DO MONDO

    To completely reload all source records and construct normalized concepts, use the
    --all and --normalize options:

        $ disease-normalizer update --all --normalize

    The Disease Normalizer will fetch the latest available data from all sources if local
    data is out-of-date. To suppress this and force usage of local files only, use the
    --use_existing flag:

        $ disease-normalizer update --all --use_existing

    \f
    :param sources: tuple of raw names of sources to update
    :param aws_instance: if true, use cloud instance
    :param db_url: URI pointing to database
    :param all_: if True, update all sources (ignore ``sources``)
    :param normalize: if True, update normalized records
    :param use_existing: if True, use most recent local data instead of fetching latest version
    :param silent: if True, suppress console output
    """  # noqa: D301
    _initialize_app()
    if len(sources) == 0 and (not all_) and (not normalize):
        click.echo(
            "Error: must provide SOURCES or at least one of --all, --normalize\n"
        )
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit(1)

    db = create_db(db_url, aws_instance)

    processed_ids = None
    try:
        from disease.etl.update import (  # noqa: PLC0415
            update_all_sources,
            update_normalized,
            update_source,
        )
    except ImportError as e:
        click.echo(
            f"Encountered ImportError: {e.msg}. Updating source data requires the optional [etl] dependency group. See the 'Full Installation' instructions in the documentation."
        )
        click.get_current_context().exit(1)
    if all_:
        processed_ids = update_all_sources(db, use_existing, silent=silent)
    elif sources:
        parsed_sources = set()
        failed_source_names = []
        for source in sources:
            try:
                parsed_sources.add(SourceName[source.upper()])
            except KeyError:
                failed_source_names.append(source)
        if len(failed_source_names) != 0:
            click.echo(f"Error: unrecognized sources: {failed_source_names}")
            click.echo(f"Valid source options are {list(SourceName)}")
            click.get_current_context().exit(1)

        working_processed_ids = set()
        for source_name in parsed_sources:
            working_processed_ids |= update_source(
                source_name, db, use_existing=use_existing, silent=silent
            )
        if len(sources) == len(SourceName):
            processed_ids = working_processed_ids

    if normalize:
        update_normalized(db, processed_ids, silent=silent)


@cli.command()
@click.option("--db_url", help=URL_DESCRIPTION)
@click.option(
    "--scope",
    type=click.Choice(list(RecordType) + list(SourceName), case_sensitive=False),
    nargs=1,
    default=RecordType.MERGER,
    help="Scope of mappings -- either an item type (merged/normalized vs base source records), or base records of an individaul source",
)
@click.option(
    "--outfile",
    "-o",
    help="Output location to write to",
    type=click.Path(path_type=Path),
)
@click.option(
    "--cancer-only",
    help="Whether to constrain mappings to just include cancers. Note: only supported by DO and MONDO records.",
    is_flag=True,
)
def dump_mappings(
    db_url: str,
    scope: RecordType | SourceName,
    outfile: Path | None,
    cancer_only: bool,
) -> None:
    """Produce JSON Lines file dump of concept referents (e.g. name/label, alias, xrefs) and
    the associated concept.

    By default, produces output for all known referents to a normalized ID. The --scope
    option can be used to constrain this either to all non-merged identity records:

        $ disease-normalizer dump-mappings --scope identity

    Or to the identity records of a specific source:

        $ disease-normalizer dump-mappings --scope ncit

    The first object in the .jsonl file will include metadata about the parameters used to
    create the document.
    """
    db = create_db(db_url, False)
    if outfile is None:
        outfile = Path() / "disease_normalizer_mappings.jsonl"
    with outfile.open("w") as f:
        meta = {
            "type": "meta",
            "created_at": datetime.datetime.now(tz=datetime.UTC).isoformat(),
            "scope": scope,
            "cancer_only": cancer_only,
        }
        f.write(json.dumps(meta))
        f.write("\n")
        for mapping in get_term_mappings(db, scope, cancer_only=cancer_only):
            f.write(json.dumps(mapping))
            f.write("\n")


if __name__ == "__main__":
    cli()



"""Provide PostgreSQL client."""

import atexit
import datetime
import logging
import os
import tarfile
import tempfile
from collections.abc import Generator
from pathlib import Path
from typing import Any, ClassVar

import psycopg
import requests
from psycopg.errors import (
    DuplicateObject,
    DuplicateTable,
    UndefinedTable,
    UniqueViolation,
)

from disease.database import AbstractDatabase, DatabaseException, DatabaseWriteException
from disease.schemas import (
    DataLicenseAttributes,
    RecordType,
    RefType,
    SourceMeta,
    SourceName,
)

_logger = logging.getLogger()


SCRIPTS_DIR = Path(__file__).parent / "postgresql"


class PostgresDatabase(AbstractDatabase):
    """Disease Normalizer database client for PostgreSQL."""

    def __init__(self, db_url: str | None = None, **db_args) -> None:
        """Initialize Postgres connection.

        .. code-block:: pycon

           >>> from disease.database.postgresql import PostgresDatabase
           >>> db = PostgresDatabase(
           >>>     user="postgres",
           >>>     password="matthew_cannon2",
           >>>     db_name="disease_normalizer"
           >>> )

        :param db_url: libpq compliant database connection URI
        :Keyword Arguments:
            * user: Postgres username
            * password: Postgres password (optional or blank if unneeded)
            * db_name: name of database to connect to
        :raise DatabaseInitializationException: if initial setup fails
        """
        if db_url:
            conninfo = db_url
        elif "DISEASE_NORM_DB_URL" in os.environ:
            conninfo = os.environ["DISEASE_NORM_DB_URL"]
        else:
            user = db_args.get("user", "postgres")
            password = db_args.get("password", "")
            db_name = db_args.get("db_name", "disease_normalizer")
            if password:
                conninfo = f"dbname={db_name} user={user} password={password}"
            else:
                conninfo = f"dbname={db_name} user={user}"

        self.conn = psycopg.connect(conninfo)
        self.initialize_db()
        self._cached_sources: dict[str, SourceMeta] = {}

        atexit.register(self.close_connection)

    _list_tables_query = b"""
    SELECT table_name FROM information_schema.tables
    WHERE table_schema = 'public'
    AND table_type = 'BASE TABLE';
    """

    def list_tables(self) -> list[str]:
        """Return names of tables in database.

        :return: Table names in database
        """
        with self.conn.cursor() as cur:
            cur.execute(self._list_tables_query)
            tables = cur.fetchall()
        return [t[0] for t in tables]

    _drop_db_query = b"""
    DROP MATERIALIZED VIEW IF EXISTS record_lookup_view;
    DROP TABLE IF EXISTS
        disease_aliases,
        disease_associations,
        disease_concepts,
        disease_labels,
        disease_merged,
        disease_sources,
        disease_xrefs;
    """

    def drop_db(self) -> None:
        """Perform complete teardown of DB. Useful for quickly resetting all data or
        reconstructing after apparent schema error. If in a protected environment,
        require confirmation.

        :raise DatabaseWriteException: if called in a protected setting with
            confirmation silenced.
        """
        try:
            if not self._check_delete_okay():
                return
        except DatabaseWriteException:
            _logger.exception("Encountered DB write error when trying to drop DB")
            raise

        with self.conn.cursor() as cur:
            cur.execute(self._drop_db_query)
            self.conn.commit()
        _logger.info("Dropped all existing disease normalizer tables.")

    def check_schema_initialized(self) -> bool:
        """Check if database schema is properly initialized.

        :return: True if DB appears to be fully initialized, False otherwise
        """
        try:
            with self.conn.cursor() as cur:
                cur.execute((SCRIPTS_DIR / "create_tables.sql").read_bytes())
        except DuplicateTable:
            self.conn.rollback()
        else:
            _logger.info("Disease table existence check failed.")
            self.conn.rollback()
            return False
        try:
            with self.conn.cursor() as cur:
                cur.execute((SCRIPTS_DIR / "add_fkeys.sql").read_bytes())
        except DuplicateObject:
            self.conn.rollback()
        else:
            _logger.info("Disease foreign key existence check failed.")
            self.conn.rollback()
            return False
        try:
            with self.conn.cursor() as cur:
                cur.execute(
                    (SCRIPTS_DIR / "create_record_lookup_view.sql").read_bytes()
                )
        except DuplicateTable:
            self.conn.rollback()
        else:
            _logger.info("Disease normalized view lookup failed.")
            self.conn.rollback()
            return False
        try:
            with self.conn.cursor() as cur:
                cur.execute((SCRIPTS_DIR / "add_indexes.sql").read_bytes())
        except DuplicateTable:
            self.conn.rollback()
        else:
            _logger.info("Disease indexes check failed.")
            self.conn.rollback()
            return False

        return True

    _check_sources_query = b"SELECT name FROM disease_sources;"
    _check_concepts_query = b"SELECT COUNT(1) FROM disease_concepts LIMIT 1;"
    _check_merged_query = b"SELECT COUNT(1) FROM disease_merged LIMIT 1;"

    def check_tables_populated(self) -> bool:
        """Perform rudimentary checks to see if tables are populated.
        Emphasis is on rudimentary -- if some fiendish element has deleted half of the
        disease aliases, this method won't pick it up. It just wants to see if a few
        critical tables have at least a small number of records.

        :return: True if queries successful, false if DB appears empty
        """
        with self.conn.cursor() as cur:
            cur.execute(self._check_sources_query)
            results = cur.fetchall()
        if len(results) < len(SourceName):
            _logger.info("Disease sources table is missing expected sources.")
            return False

        with self.conn.cursor() as cur:
            cur.execute(self._check_concepts_query)
            result = cur.fetchone()
        if not result or result[0] < 1:
            _logger.info("Disease records table is empty.")
            return False

        with self.conn.cursor() as cur:
            cur.execute(self._check_merged_query)
            result = cur.fetchone()
        if not result or result[0] < 1:
            _logger.info("Normalized disease records table is empty.")
            return False

        return True

    def initialize_db(self) -> None:
        """Check if DB is set up. If not, create tables/indexes/views."""
        if not self.check_schema_initialized():
            self.drop_db()
            self._create_tables()
            self._create_views()
            self._add_indexes()

    def _create_views(self) -> None:
        """Create materialized views."""
        create_view_query = (SCRIPTS_DIR / "create_record_lookup_view.sql").read_bytes()
        with self.conn.cursor() as cur:
            cur.execute(create_view_query)
            self.conn.commit()

    _refresh_views_query = b"REFRESH MATERIALIZED VIEW record_lookup_view;"

    def _refresh_views(self) -> None:
        """Update materialized views.

        Not responsible for ensuring existence of views. Calling functions should
        either check beforehand or catch psycopg.UndefinedTable.
        """
        with self.conn.cursor() as cur:
            cur.execute(self._refresh_views_query)
            self.conn.commit()

    def _add_fkeys(self) -> None:
        """Add fkey relationships."""
        add_fkey_query = (SCRIPTS_DIR / "add_fkeys.sql").read_bytes()
        with self.conn.cursor() as cur:
            cur.execute(add_fkey_query)
            self.conn.commit()

    def _drop_fkeys(self) -> None:
        """Drop fkey relationships."""
        drop_fkey_query = (SCRIPTS_DIR / "drop_fkeys.sql").read_bytes()
        with self.conn.cursor() as cur:
            cur.execute(drop_fkey_query)
            self.conn.commit()

    def _add_indexes(self) -> None:
        """Create core search indexes."""
        add_indexes_query = (SCRIPTS_DIR / "add_indexes.sql").read_bytes()
        with self.conn.cursor() as cur:
            cur.execute(add_indexes_query)
            self.conn.commit()

    def _drop_indexes(self) -> None:
        """Drop all custom indexes."""
        drop_indexes_query = (SCRIPTS_DIR / "drop_indexes.sql").read_bytes()
        with self.conn.cursor() as cur:
            cur.execute(drop_indexes_query)
            self.conn.commit()

    def _create_tables(self) -> None:
        """Create all tables, indexes, and views."""
        _logger.debug("Creating new disease normalizer tables.")
        tables_query = (SCRIPTS_DIR / "create_tables.sql").read_bytes()

        with self.conn.cursor() as cur:
            cur.execute(tables_query)
            self.conn.commit()

    _source_metadata_query = b"SELECT * FROM disease_sources WHERE name = %s;"

    def get_source_metadata(self, src_name: str | SourceName) -> SourceMeta | None:
        """Get license, versioning, data lookup, etc information for a source.

        :param src_name: name of the source to get data for
        :return: source metadata, if lookup is successful
        """
        if isinstance(src_name, SourceName):
            src_name = src_name.value

        if src_name in self._cached_sources:
            return self._cached_sources[src_name]

        with self.conn.cursor() as cur:
            cur.execute(self._source_metadata_query, [src_name])
            metadata_result = cur.fetchone()
            if not metadata_result:
                return None
            metadata = SourceMeta(
                data_license=metadata_result[1],
                data_license_url=metadata_result[2],
                version=metadata_result[3],
                data_url=metadata_result[4],
                rdp_url=metadata_result[5],
                data_license_attributes=DataLicenseAttributes(
                    non_commercial=metadata_result[6],
                    attribution=metadata_result[7],
                    share_alike=metadata_result[8],
                ),
            )
            self._cached_sources[src_name] = metadata
            return metadata

    _record_query = b"SELECT * FROM record_lookup_view WHERE lower(concept_id) = %s;"

    def _format_source_record(self, source_row: tuple) -> dict:
        """Restructure row from disease_concepts table as source record result object.

        :param source_row: result tuple from psycopg
        :return: reformatted dictionary keying disease properties to row values
        """
        disease_record = {
            "concept_id": source_row[0],
            "label": source_row[1],
            "aliases": source_row[2],
            "associated_with": source_row[3],
            "xrefs": source_row[4],
            "src_name": source_row[5],
            "merge_ref": source_row[6],
            "pediatric_disease": source_row[7],
            "oncologic_disease": source_row[8],
            "item_type": RecordType.IDENTITY.value,
        }
        return {k: v for k, v in disease_record.items() if v}

    def _get_record(self, concept_id: str) -> dict | None:
        """Retrieve non-merged record. The query is pretty different, so this method
        is broken out for PostgreSQL.

        :param concept_id: ID of concept to get
        :return: complete record object if successful
        """
        concept_id_param = concept_id.lower()

        with self.conn.cursor() as cur:
            cur.execute(self._record_query, [concept_id_param])
            result = cur.fetchone()
        if not result:
            return None

        return self._format_source_record(result)

    _merged_record_query = b"SELECT * FROM disease_merged WHERE lower(concept_id) = %s;"

    def _format_merged_record(self, merged_row: tuple) -> dict:
        """Restructure row from disease_merged table as normalized result object.

        :param merged_row: result tuple from psycopg
        :return: reformatted dictionary keying normalized disease properties to row values
        """
        merged_record = {
            "concept_id": merged_row[0],
            "label": merged_row[1],
            "aliases": merged_row[2],
            "associated_with": merged_row[3],
            "xrefs": merged_row[4],
            "pediatric_disease": merged_row[5],
            "oncologic_disease": merged_row[6],
            "item_type": RecordType.MERGER.value,
        }
        return {k: v for k, v in merged_record.items() if v}

    def _get_merged_record(self, concept_id: str) -> dict | None:
        """Retrieve normalized record from DB.

        :param concept_id: normalized ID for the merged record
        :return: normalized record if successful
        """
        concept_id = concept_id.lower()
        with self.conn.cursor() as cur:
            cur.execute(self._merged_record_query, [concept_id])
            result = cur.fetchone()
        if not result:
            return None

        return self._format_merged_record(result)

    def get_record_by_id(
        self,
        concept_id: str,
        case_sensitive: bool = True,  # noqa: ARG002
        merge: bool = False,
    ) -> dict | None:
        """Fetch record corresponding to provided concept ID

        :param str concept_id: concept ID for disease record
        :param bool case_sensitive: not used by this implementation -- lookups use
           case-insensitive index
        :param bool merge: if true, look for merged record; look for identity record
           otherwise.
        :return: complete disease record, if match is found; None otherwise
        """
        if merge:
            return self._get_merged_record(concept_id)
        return self._get_record(concept_id)

    _ref_types_query: ClassVar[dict] = {
        RefType.LABEL: b"SELECT concept_id FROM disease_labels WHERE lower(label) = %s;",
        RefType.ALIASES: b"SELECT concept_id FROM disease_aliases WHERE lower(alias) = %s;",
        RefType.XREFS: "SELECT concept_id FROM disease_xrefs WHERE lower(xref) = %s;",
        RefType.ASSOCIATED_WITH: "SELECT concept_id FROM disease_associations WHERE lower(associated_with) = %s;",
    }

    def get_refs_by_type(self, search_term: str, ref_type: RefType) -> list[str]:
        """Retrieve concept IDs for records matching the user's query. Other methods
        are responsible for actually retrieving full records.

        :param search_term: string to match against
        :param ref_type: type of match to look for.
        :return: list of associated concept IDs. Empty if lookup fails.
        """
        query = self._ref_types_query.get(ref_type)
        if not query:
            msg = "invalid reference type"
            raise ValueError(msg)

        with self.conn.cursor() as cur:
            cur.execute(query, (search_term.lower(),))
            concept_ids = cur.fetchall()
        if concept_ids:
            return [i[0] for i in concept_ids]
        return []

    _source_ids_query = b"""
    SELECT concept_id FROM disease_concepts dc
    LEFT JOIN disease_sources ds ON ds.name = dc.source
    WHERE ds.name = %s;
    """
    _all_ids_query = b"SELECT concept_id FROM disease_concepts;"

    def get_all_concept_ids(self, source: SourceName | None = None) -> set[str]:
        """Retrieve concept IDs for use in generating normalized records.

        :param source: optionally, just get all IDs for a specific source
        :return: Set of concept IDs as strings.
        """
        if source is not None:
            ids_query = self._source_ids_query
        else:
            ids_query = self._all_ids_query
        with self.conn.cursor() as cur:
            if source is None:
                cur.execute(ids_query)
            else:
                cur.execute(ids_query, (source,))
            ids_tuple = cur.fetchall()
        return {i[0] for i in ids_tuple}

    _get_all_normalized_records_query = b"SELECT * FROM disease_merged;"
    _get_all_unmerged_source_records_query = (
        b"SELECT * FROM record_lookup_view WHERE merge_ref IS NULL;"
    )
    _get_all_source_records_query = b"SELECT * FROM record_lookup_view;"

    def get_all_records(self, record_type: RecordType) -> Generator[dict, None, None]:
        """Retrieve all source or normalized records. Either return all source records,
        or all records that qualify as "normalized" (i.e., merged groups + source
        records that are otherwise ungrouped).

        For example,

        .. code-block:: pycon

           >>> from disease.database import create_db
           >>> from disease.schemas import RecordType
           >>> db = create_db()
           >>> for record in db.get_all_records(RecordType.MERGER):
           >>>     pass  # do something

        Unlike DynamoDB, merged records are stored in a separate table from source
        records. As a result, when fetching all normalized records, merged records are
        return first, and iteration continues with all source records that don't
        belong to a normalized concept group.

        :param record_type: type of result to return
        :return: Generator that lazily provides records as they are retrieved
        """
        batch_size = 500

        if record_type == RecordType.MERGER:
            with self.conn.cursor() as cur:
                results = cur.execute(self._get_all_normalized_records_query)
                fetched = results.fetchmany(batch_size)
                while fetched:
                    for row in fetched:
                        yield self._format_merged_record(row)
                    fetched = results.fetchmany(batch_size)
            with self.conn.cursor() as cur:
                results = cur.execute(self._get_all_unmerged_source_records_query)
                fetched = results.fetchmany(batch_size)
                while fetched:
                    for result in fetched:
                        yield self._format_source_record(result)
                    fetched = results.fetchmany(batch_size)
        else:
            with self.conn.cursor() as cur:
                results = cur.execute(self._get_all_source_records_query)
                fetched = results.fetchmany(batch_size)
                while fetched:
                    for result in fetched:
                        yield self._format_source_record(result)
                    fetched = results.fetchmany(batch_size)

    _add_source_metadata_query = b"""
    INSERT INTO disease_sources(
        name, data_license, data_license_url, version, data_url, rdp_url,
        data_license_nc, data_license_attr, data_license_sa
    )
    VALUES ( %s, %s, %s, %s, %s, %s, %s, %s, %s );
    """

    def add_source_metadata(self, src_name: SourceName, meta: SourceMeta) -> None:
        """Add new source metadata entry.

        :param src_name: name of source
        :param meta: known source attributes
        :raise DatabaseWriteException: if write fails
        """
        with self.conn.cursor() as cur:
            cur.execute(
                self._add_source_metadata_query,
                [
                    src_name.value,
                    meta.data_license,
                    meta.data_license_url,
                    meta.version,
                    meta.data_url,
                    meta.rdp_url,
                    meta.data_license_attributes.non_commercial,
                    meta.data_license_attributes.attribution,
                    meta.data_license_attributes.share_alike,
                ],
            )
        self.conn.commit()

    _add_record_query = b"""
        INSERT INTO disease_concepts (concept_id, source, pediatric_disease, oncologic_disease)
        VALUES (%s, %s, %s, %s);
    """
    _insert_label_query = (
        b"INSERT INTO disease_labels (label, concept_id) VALUES (%s, %s)"
    )
    _insert_alias_query = (
        b"INSERT INTO disease_aliases (alias, concept_id) VALUES (%s, %s)"
    )
    _insert_xref_query = b"INSERT INTO disease_xrefs (xref, concept_id) VALUES (%s, %s)"
    _insert_assoc_query = b"INSERT INTO disease_associations (associated_with, concept_id) VALUES (%s, %s)"

    def add_record(self, record: dict, src_name: SourceName) -> None:
        """Add new record to database.

        :param record: record to upload
        :param src_name: name of source for record. Not used by PostgreSQL instance.
        """
        concept_id = record["concept_id"]
        with self.conn.cursor() as cur:
            try:
                cur.execute(
                    self._add_record_query,
                    [
                        concept_id,
                        src_name.value,
                        record.get("pediatric_disease"),
                        record.get("oncologic_disease"),
                    ],
                )
                cur.execute(self._insert_label_query, [record["label"], concept_id])
                for a in record.get("aliases", []):
                    cur.execute(self._insert_alias_query, [a, concept_id])
                for x in record.get("xrefs", []):
                    cur.execute(self._insert_xref_query, [x, concept_id])
                for a in record.get("associated_with", []):
                    cur.execute(self._insert_assoc_query, [a, concept_id])
                self.conn.commit()
            except UniqueViolation:
                _logger.error("Record with ID %s already exists", concept_id)  # noqa: TRY400
                self.conn.rollback()

    _add_merged_record_query = b"""
    INSERT INTO disease_merged (
        concept_id, label, aliases, associated_with, xrefs, pediatric_disease, oncologic_disease
    )
    VALUES (%s, %s, %s, %s, %s, %s, %s);
    """

    def add_merged_record(self, record: dict) -> None:
        """Add merged record to database.

        :param record: merged record to add
        """
        with self.conn.cursor() as cur:
            cur.execute(
                self._add_merged_record_query,
                [
                    record["concept_id"],
                    record["label"],
                    record.get("aliases"),
                    record.get("associated_with"),
                    record.get("xrefs"),
                    record.get("pediatric_disease"),
                    record.get("oncologic_disease"),
                ],
            )
            self.conn.commit()

    _update_merge_ref_query = b"""
    UPDATE disease_concepts
    SET merge_ref = %(merge_ref)s
    WHERE concept_id = %(concept_id)s;
    """

    def update_merge_ref(self, concept_id: str, merge_ref: Any) -> None:  # noqa: ANN401
        """Update the merged record reference of an individual record to a new value.

        :param concept_id: record to update
        :param merge_ref: new ref value
        :raise DatabaseWriteException: if attempting to update non-existent record
        """
        with self.conn.cursor() as cur:
            cur.execute(
                self._update_merge_ref_query,
                {"merge_ref": merge_ref, "concept_id": concept_id},
            )
            row_count = cur.rowcount
            self.conn.commit()

        # UPDATE will fail silently unless we check the # of affected rows
        if row_count < 1:
            msg = f"No such record exists for primary key {concept_id}"
            raise DatabaseWriteException(msg)

    def delete_normalized_concepts(self) -> None:
        """Remove merged records from the database. Use when performing a new update
        of normalized data.

        It would be faster to drop the entire table and do a cascading delete onto the
        merge_ref column in disease_concepts, but that requires an exclusive access lock
        on the DB, which can be annoying (ie you couldn't have multiple processes
        accessing it, or PgAdmin, etc...). Instead, we'll take down each merge table
        dependency and rebuild afterwards.

        :raise DatabaseReadException: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteException: if deletion call fails
        """
        with self.conn.cursor() as cur:
            cur.execute((SCRIPTS_DIR / "delete_normalized_concepts.sql").read_bytes())
        self.conn.commit()

    _drop_aliases_query = b"""
    DELETE FROM disease_aliases WHERE id IN (
        SELECT da.id FROM disease_aliases da LEFT JOIN disease_concepts dc
            ON dc.concept_id = da.concept_id
        WHERE dc.source = %s
    );
    """
    _drop_associations_query = b"""
    DELETE FROM disease_associations WHERE id IN (
        SELECT da.id FROM disease_associations da LEFT JOIN disease_concepts dc
            ON dc.concept_id = da.concept_id
        WHERE dc.source = %s
    );
    """
    _drop_labels_query = b"""
    DELETE FROM disease_labels WHERE id IN (
        SELECT ds.id FROM disease_labels ds LEFT JOIN disease_concepts dc
            ON dc.concept_id = ds.concept_id
        WHERE dc.source = %s
    );
    """
    _drop_xrefs_query = b"""
    DELETE FROM disease_xrefs WHERE id IN (
        SELECT dx.id FROM disease_xrefs dx LEFT JOIN disease_concepts dc
            ON dc.concept_id = dx.concept_id
        WHERE dc.source = %s
    );
    """
    _drop_concepts_query = b"DELETE FROM disease_concepts WHERE source = %s;"
    _drop_source_query = b"DELETE FROM disease_sources ds WHERE ds.name = %s;"

    def delete_source(self, src_name: SourceName) -> None:
        """Delete all data for a source. Use when updating source data.

        All of the foreign key relations make deletes *extremely* slow, so this method
        drops and then re-adds them once deletes are finished. This makes it a little
        brittle, and it'd be nice to revisit in the future to perform as a single
        atomic transaction.

        Refreshing the materialized view at the end might be redundant, because
        this method will almost always be called right before more data is written,
        but it's probably necessary just in case that doesn't happen.

        :param src_name: name of source to delete
        :raise DatabaseWriteException: if deletion call fails
        """
        with self.conn.cursor() as cur:
            cur.execute(self._drop_aliases_query, [src_name.value])
            cur.execute(self._drop_associations_query, [src_name.value])
            cur.execute(self._drop_labels_query, [src_name.value])
            cur.execute(self._drop_xrefs_query, [src_name.value])
        self._drop_fkeys()
        self._drop_indexes()

        with self.conn.cursor() as cur:
            cur.execute(self._drop_concepts_query, [src_name.value])
            cur.execute(self._drop_source_query, [src_name.value])
            self.conn.commit()

        self._add_fkeys()
        self._add_indexes()
        self._refresh_views()

    def complete_write_transaction(self) -> None:
        """Conclude transaction or batch writing if relevant."""
        if not self.conn.closed:
            try:
                self._refresh_views()
            except UndefinedTable:
                self.conn.rollback()

    def close_connection(self) -> None:
        """Perform any manual connection closure procedures if necessary."""
        if not self.conn.closed:
            self.conn.commit()
            self.conn.close()

    def load_from_remote(self, url: str | None) -> None:
        """Load DB from remote dump. Warning: Deletes all existing data. If not
        passed as an argument, will try to grab latest release from VICC S3 bucket.

        :param url: location of .tar.gz file created from output of pg_dump
        :raise DatabaseException: if unable to retrieve file from URL or if psql
            command fails
        """
        if not url:
            url = "https://vicc-normalizers.s3.us-east-2.amazonaws.com/disease_normalization/postgresql/disease_norm_latest.sql.tar.gz"
        with tempfile.TemporaryDirectory() as tempdir:
            tempdir_path = Path(tempdir)
            temp_tarfile = tempdir_path / "disease_norm_latest.tar.gz"
            with requests.get(url, stream=True, timeout=30) as r:
                try:
                    r.raise_for_status()
                except requests.HTTPError as e:
                    msg = f"Unable to retrieve PostgreSQL dump file from {url}"
                    raise DatabaseException(msg) from e
                with temp_tarfile.open("wb") as h:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:
                            h.write(chunk)
            tar = tarfile.open(temp_tarfile, "r:gz")  # noqa: SIM115
            tar_dump_file = next(
                f for f in tar.getmembers() if f.name.startswith("disease_norm_")
            )
            tar.extractall(path=tempdir_path, members=[tar_dump_file])  # noqa: S202 RUF100
            dump_file = tempdir_path / tar_dump_file.name

            if self.conn.info.password:
                pw_param = f"-W {self.conn.info.password}"
            else:
                pw_param = "-w"

            self.drop_db()
            system_call = f"psql -d {self.conn.info.dbname} -U {self.conn.info.user} {pw_param} -f {dump_file.absolute()}"
            result = os.system(system_call)  # noqa: S605
        if result != 0:
            msg = f"System call '{result}' returned failing exit code."
            raise DatabaseException(msg)

    def export_db(self, output_directory: Path) -> None:
        """Dump DB to specified location.

        :param export_location: path to directory to save DB dump in
        :return: Nothing, but saves results of pg_dump to file named
            ``disease_norm_<date and time>.sql``
        :raise ValueError: if output directory isn't a directory or doesn't exist
        :raise DatabaseException: if psql call fails
        """
        if not output_directory.is_dir() or not output_directory.exists():
            msg = (
                f"Output location {output_directory} isn't a directory or doesn't exist"
            )
            raise ValueError(msg)
        now = datetime.datetime.now(tz=datetime.UTC).strftime("%Y%m%d%H%M%S")
        output_location = output_directory / f"disease_norm_{now}.sql"
        user = self.conn.info.user
        host = self.conn.info.host
        port = self.conn.info.port
        database_name = self.conn.info.dbname
        pw_param = f"-W {self.conn.info.password}" if self.conn.info.password else "-w"

        system_call = f"pg_dump -E UTF8 -f {output_location} -U {user} {pw_param} -h {host} -p {port} {database_name}"
        _logger.debug("Executing DB dump system call: %s", system_call)
        result = os.system(system_call)  # noqa: S605
        if result != 0:
            msg = f"System call '{system_call}' returned failing exit code."
            raise DatabaseException(msg)



"""Read and provide runtime configuration."""

from functools import cache

from pydantic_settings import BaseSettings, SettingsConfigDict

from disease.schemas import ServiceEnvironment


class Settings(BaseSettings):
    """Create app settings

    This is not a singleton, so every new call to this class will re-compute
    configuration settings, defaults, etc.
    """

    model_config = SettingsConfigDict(
        env_prefix="DISEASE_NORM_",
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )

    env: ServiceEnvironment = ServiceEnvironment.DEV
    debug: bool = False
    test: bool = False
    db_url: str = "http://localhost:8000"


@cache
def get_config() -> Settings:
    """Get runtime configuration.

    This function is cached, so the config object only gets created/calculated once.

    :return: Settings instance
    """
    return Settings()



"""Get NCIt disease data."""

import logging
import re

import owlready2 as owl
from tqdm import tqdm

from disease.etl.base import OWLBase
from disease.schemas import NamespacePrefix, SourceMeta, SourceName

_logger = logging.getLogger(__name__)

icdo_re = re.compile("[0-9]+/[0-9]+")


class NCIt(OWLBase):
    """Gather and load data from NCIt."""

    def _load_meta(self) -> None:
        """Load metadata"""
        metadata = SourceMeta(
            data_license="CC BY 4.0",
            data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",  # F401
            version=self._version,
            data_url="https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/",
            rdp_url="http://reusabledata.org/ncit.html",
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": True,
            },
        )
        self._database.add_source_metadata(self._src_name, metadata)

    def _get_disease_classes(self) -> set[str]:
        """Get all nodes with semantic_type 'Neoplastic Process' or 'Disease
        or Syndrome'.

        :return: uq_nodes with additions from above types added
        :rtype: Set[str]
        """
        graph = owl.default_world.as_rdflib_graph()
        p106 = "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P106"
        neopl = self._get_by_property_value(p106, "Neoplastic Process", graph)
        dos = self._get_by_property_value(p106, "Disease or Syndrome", graph)
        p310 = "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P310"
        retired = self._get_by_property_value(p310, "Retired_Concept", graph)
        c22187 = "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#C22187"
        nonhuman = self._get_subclasses(c22187, graph)
        return neopl.union(dos) - retired - nonhuman

    def _transform_data(self) -> None:
        """Get data from file and construct object for loading."""
        ncit = owl.get_ontology(self._data_file.absolute().as_uri()).load()
        disease_uris = self._get_disease_classes()
        for uri in tqdm(disease_uris, ncols=80, disable=self._silent):
            disease_class = ncit.search(iri=uri)[0]
            concept_id = f"{NamespacePrefix.NCIT.value}:{disease_class.name}"
            if disease_class.P108:
                label = disease_class.P108.first()
            else:
                _logger.warning("No label for concept %s", concept_id)
                continue
            aliases = [a for a in disease_class.P90 if a != label]

            associated_with = []
            if disease_class.P207:
                associated_with.append(
                    f"{NamespacePrefix.UMLS.value}:{disease_class.P207.first()}"
                )
            maps_to = disease_class.P375
            if maps_to:
                icdo_list = list(filter(lambda s: icdo_re.match(s), maps_to))
                if len(icdo_list) == 1:
                    associated_with.append(
                        f"{NamespacePrefix.ICDO.value}:{icdo_list[0]}"
                    )
            imdrf = disease_class.hasDbXref
            if imdrf:
                associated_with.append(
                    f"{NamespacePrefix.IMDRF.value}:{imdrf[0].split(':')[1]}"
                )

            disease = {
                "concept_id": concept_id,
                "src_name": SourceName.NCIT.value,
                "label": label,
                "aliases": aliases,
                "associated_with": associated_with,
            }
            self._load_disease(disease)



"""Test retrieval of Disease Ontology source data."""

import pytest

from disease.etl.do import DO
from disease.schemas import Disease, MatchType


@pytest.fixture(scope="module")
def do(test_source):
    """Provide test DO query endpoint"""
    return test_source(DO)


@pytest.fixture(scope="module")
def neuroblastoma():
    """Build neuroblastoma fixture."""
    return Disease(
        concept_id="DOID:769",
        label="neuroblastoma",
        xrefs=["ncit:C3270"],
        associated_with=[
            "efo:0000621",
            "gard:7185",
            "icdo:9500/3",
            "mesh:D009447",
            "orphanet:635",
            "umls:C0027819",
        ],
        aliases=[],
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def pediatric_liposarcoma():
    """Create test fixture for pediatric liposarcoma."""
    return Disease(
        concept_id="DOID:5695",
        label="childhood liposarcoma",
        xrefs=["ncit:C8091"],
        associated_with=["umls:C0279984"],
        aliases=["pediatric liposarcoma"],
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def richter():
    """Create test fixture for Richter's Syndrome."""
    return Disease(
        concept_id="DOID:1703",
        label="Richter's syndrome",
        aliases=["Richter syndrome"],
        xrefs=["ncit:C35424"],
        associated_with=["umls:C0349631", "gard:7578", "icd10.cm:C91.1"],
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def juvenile_testicular_granulosa_cell():
    """Fixture for testing blocklist function"""
    return Disease(
        concept_id="DOID:6032",
        label="juvenile type testicular granulosa cell tumor",
        aliases=[
            "Juvenile granulosa cell tumor",
            "Juvenile granulosa cell tumour",
            "Juvenile type Granulosa cell tumor",
            "Juvenile type Granulosa cell tumour",
            "juvenile type granulosa cell neoplasm",
            "juvenile type testicular granulosa cell tumour",
        ],
        xrefs=["ncit:C39947"],
        associated_with=["umls:C1515285", "umls:C0334403"],
        oncologic_disease=True,
    )


def test_concept_id_match(
    do, neuroblastoma, pediatric_liposarcoma, richter, compare_response
):
    """Test that concept ID search resolves to correct record"""
    response = do.search("DOID:769")
    compare_response(response, MatchType.CONCEPT_ID, neuroblastoma)

    response = do.search("doid:5695")
    compare_response(response, MatchType.CONCEPT_ID, pediatric_liposarcoma)

    response = do.search("DOid:1703")
    compare_response(response, MatchType.CONCEPT_ID, richter)

    response = do.search("5695")
    assert response.match_type == MatchType.NO_MATCH


def test_label_match(do, neuroblastoma, pediatric_liposarcoma, compare_response):
    """Test that label searches resolve to correct records."""
    response = do.search("childhood liposarcoma")
    compare_response(response, MatchType.LABEL, pediatric_liposarcoma)

    response = do.search("NEUROBLASTOMA")
    compare_response(response, MatchType.LABEL, neuroblastoma)


def test_alias_match(do, richter, compare_response):
    """Test that alias searches resolve to correct records."""
    response = do.search("Richter Syndrome")
    compare_response(response, MatchType.ALIAS, richter)


def test_xref_match(
    do,
    neuroblastoma: Disease,
    pediatric_liposarcoma: Disease,
    juvenile_testicular_granulosa_cell: Disease,
    compare_response,
):
    """Test that xref search resolves to correct records."""
    response = do.search("ncit:C3270")
    compare_response(response, MatchType.XREF, neuroblastoma)

    response = do.search("NCIT:C8091")
    compare_response(response, MatchType.XREF, pediatric_liposarcoma)

    response = do.search("ncit:C39947")
    compare_response(response, MatchType.XREF, juvenile_testicular_granulosa_cell)

    response = do.search("ncit:C4207")  # blocked by blocklist
    assert response.match_type == MatchType.NO_MATCH


def test_associated_with_match(
    do, neuroblastoma, pediatric_liposarcoma, richter, compare_response
):
    """Test that associated_with search resolves to correct records."""
    response = do.search("umls:c0027819")
    compare_response(response, MatchType.ASSOCIATED_WITH, neuroblastoma)

    response = do.search("umls:C0279984")
    compare_response(response, MatchType.ASSOCIATED_WITH, pediatric_liposarcoma)

    response = do.search("icd10.cm:c91.1")
    compare_response(response, MatchType.ASSOCIATED_WITH, richter)



"""Test MONDO ETL methods."""

import re

import pytest

from disease.etl.mondo import Mondo
from disease.schemas import Disease, MatchType


@pytest.fixture(scope="module")
def mondo(test_source):
    """Build Mondo ETL test fixture."""
    return test_source(Mondo)


@pytest.fixture(scope="module")
def neuroblastoma():
    """Construct a test fixture for neuroblastoma"""
    return Disease(
        concept_id="mondo:0005072",
        label="neuroblastoma",
        aliases=[
            "NB",
            "neuroblastoma",
            "neural Crest tumor, malignant",
            "neuroblastoma (Schwannian Stroma-poor)",
            "neuroblastoma, malignant",
        ],
        xrefs=[
            "ncit:C3270",
            "DOID:769",
            "oncotree:NBL",
        ],
        associated_with=[
            "orphanet:635",
            "umls:C0027819",
            "efo:0000621",
            "mesh:D009447",
            "medgen:18012",
        ],
        pediatric_disease=None,
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def richter_syndrome():
    """Construct a test fixture for Richter Syndrome"""
    return Disease(
        concept_id="mondo:0002083",
        label="Richter syndrome",
        aliases=[
            "Richter syndrome",
            "Richter's syndrome",
            "Richter transformation",
            "Richter's transformation",
        ],
        xrefs=["ncit:C35424", "DOID:1703"],
        associated_with=["umls:C0349631", "medgen:91159"],
        pediatric_disease=None,
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def pediatric_liposarcoma():
    """Construct a test fixture for pediatric liposarcoma. Tests the
    pediatric flag.
    """
    return Disease(
        concept_id="mondo:0003587",
        label="pediatric liposarcoma",
        aliases=[
            "pediatric liposarcoma",
            "childhood liposarcoma",
            "liposarcoma",
        ],
        xrefs=["DOID:5695", "ncit:C8091"],
        associated_with=["umls:C0279984", "medgen:83580"],
        pediatric_disease=True,
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def cystic_teratoma_adult():
    """Construct a test fixture for adult cystic teratoma. Tests the
    pediatric flag.
    """
    return Disease(
        concept_id="mondo:0004099",
        label="adult cystic teratoma",
        aliases=["cystic teratoma of adults", "adult cystic teratoma"],
        pediatric_disease=None,
        xrefs=["ncit:C9012", "DOID:7079"],
        associated_with=["umls:C1368888", "medgen:235084"],
        oncologic_disease=True,
    )


@pytest.fixture(scope="module")
def nsclc():
    """Construct a test fixture for non small cell lung cancer."""
    return Disease(
        concept_id="mondo:0005233",
        label="non-small cell lung carcinoma",
        aliases=[
            "NSCLC - non-small cell lung cancer",
            "non-small cell lung carcinoma (disease)",
            "non-small cell carcinoma of lung",
            "non-small cell carcinoma of the lung",
            "non-small cell cancer of lung",
            "non-small cell lung cancer",
            "non-small cell cancer of the lung",
            "NSCLC",
            "non-small cell lung carcinoma",
        ],
        xrefs=["ncit:C2926", "oncotree:NSCLC", "DOID:3908"],
        associated_with=[
            "mesh:D002289",
            "umls:C0007131",
            "efo:0003060",
            "medgen:40104",
        ],
        oncologic_disease=True,
    )


def test_neuroblastoma(mondo, neuroblastoma, compare_response, compare_records):
    response = mondo.search("MONDO:0005072")
    compare_response(response, MatchType.CONCEPT_ID, neuroblastoma)

    response = mondo.search("Neuroblastoma")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)

    response = mondo.search("NEUROBLASTOMA")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)

    response = mondo.search("neuroblastoma, malignant")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)

    response = mondo.search("DOID:769")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)

    response = mondo.search("orphanet:635")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)

    response = mondo.search("Neuroblastoma, Malignant")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)

    response = mondo.search("neuroblastoma (Schwannian Stroma-poor)")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, neuroblastoma)


def test_richter_syndrome(mondo, richter_syndrome, compare_records, compare_response):
    response = mondo.search("mondo:0002083")
    compare_response(response, MatchType.CONCEPT_ID, richter_syndrome)

    response = mondo.search("richter syndrome")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, richter_syndrome)

    response = mondo.search("umls:C0349631")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, richter_syndrome)

    response = mondo.search("RICHTER TRANSFORMATION")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, richter_syndrome)

    response = mondo.search("Richter's transformation")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, richter_syndrome)

    response = mondo.search("Richter's syndrome")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, richter_syndrome)


def test_pediatric_liposarcoma(
    mondo, pediatric_liposarcoma, compare_records, compare_response
):
    response = mondo.search("mondo:0003587")
    compare_response(response, MatchType.CONCEPT_ID, pediatric_liposarcoma)

    response = mondo.search("pediatric liposarcoma")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, pediatric_liposarcoma)

    response = mondo.search("ncit:c8091")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, pediatric_liposarcoma)

    response = mondo.search("UMLS:C0279984")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, pediatric_liposarcoma)


def test_cystic_teratoma_adult(
    mondo,
    cystic_teratoma_adult,
    compare_response,
    compare_records,
):
    response = mondo.search("mondo:0004099")
    compare_response(response, MatchType.CONCEPT_ID, cystic_teratoma_adult)

    response = mondo.search("Adult Cystic Teratoma")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, cystic_teratoma_adult)


def test_nsclc(
    mondo,
    nsclc,
    compare_response,
    compare_records,
):
    response = mondo.search("mondo:0005233")
    compare_response(response, MatchType.CONCEPT_ID, nsclc)

    response = mondo.search("oncotree:NSCLC")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, nsclc)

    response = mondo.search("ncit:C2926")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    actual_disease = response.records[0]
    compare_records(actual_disease, nsclc)


def test_meta(mondo):
    """Test that meta field is correct."""
    response = mondo.search("neuroblastoma")
    assert response.source_meta_.data_license == "CC BY 4.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by/4.0/legalcode"
    )
    assert re.match(r"\d{4}\d{2}\d{2}", response.source_meta_.version)
    assert (
        response.source_meta_.data_url
        == "https://mondo.monarchinitiative.org/pages/download/"
    )
    assert response.source_meta_.rdp_url == "http://reusabledata.org/monarch.html"
    assert not response.source_meta_.data_license_attributes.non_commercial
    assert not response.source_meta_.data_license_attributes.share_alike
    assert response.source_meta_.data_license_attributes.attribution



"""Construct test data for NCIt source."""

import xml.etree.ElementTree as XET  # noqa: N814
from collections.abc import Generator
from pathlib import Path

import lxml.etree as ET  # noqa: N812
import owlready2 as owl
import xmlformatter

from disease.database import create_db
from disease.etl import NCIt

# define captured ids in `test_classes` variable

NCIT_PREFIX = "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl"
OWL_PREFIX = "{http://www.w3.org/2002/07/owl#}"
RDFS_PREFIX = "{http://www.w3.org/2000/01/rdf-schema#}"
RDF_PREFIX = "{http://www.w3.org/1999/02/22-rdf-syntax-ns#}"

ANNOTATION_PROPERTY_TAG = f"{OWL_PREFIX}AnnotationProperty"
DESCRIPTION_TAG = f"{RDF_PREFIX}Description"
ABOUT_ATTRIB = f"{RDF_PREFIX}about"
AXIOM_TAG = f"{OWL_PREFIX}Axiom"
DATATYPE_TAG = f"{RDFS_PREFIX}Datatype"
OBJECT_PROPERTY_TAG = f"{OWL_PREFIX}ObjectProperty"
CLASS_TAG = f"{OWL_PREFIX}Class"


ncit = NCIt(create_db())
ncit._extract_data()
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "ncit"
outfile_path = TEST_DATA_DIR / ncit._data_file.name
outfile_path.parent.mkdir(exist_ok=True)


def ncit_parser() -> Generator:
    """Get unique XML elements."""
    context = iter(
        ET.iterparse(ncit._data_file, events=("start", "end"), huge_tree=True)
    )
    for event, elem in context:
        if event == "start":
            yield elem


parser = ncit_parser()

# make root element
root = next(parser)
nsmap = root.nsmap
new_root = ET.Element(f"{RDF_PREFIX}RDF", nsmap=nsmap)
new_root.set(
    "{http://www.w3.org/XML/1998/namespace}base",
    "http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl",
)

# add ontology element
element = next(parser)
new_root.append(element)

# get annotation property elements
element = next(parser)
while element.tag != ANNOTATION_PROPERTY_TAG:
    element = next(parser)

while element.tag != DESCRIPTION_TAG:
    new_root.append(element)
    element = next(parser)
    while element.tag != ANNOTATION_PROPERTY_TAG and element.tag != DESCRIPTION_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get description elements
descriptions = {}
while element.tag != DATATYPE_TAG:
    new_root.append(element)
    element = next(parser)
    while element.tag != DESCRIPTION_TAG and element.tag != DATATYPE_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get datatype elements
while element.tag != OBJECT_PROPERTY_TAG:
    new_root.append(element)
    element = next(parser)
    while (
        element.tag != DATATYPE_TAG or len(element.attrib) == 0
    ) and element.tag != OBJECT_PROPERTY_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get object property elements
while element.tag != CLASS_TAG:
    new_root.append(element)
    element = next(parser)
    while element.tag != OBJECT_PROPERTY_TAG and element.tag != CLASS_TAG:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        element = next(parser)

# get classes
onto = owl.get_ontology(ncit._data_file.absolute().as_uri())
onto.load()
test_classes = {
    onto.C2926,
    onto.C3270,
    onto.C8091,
    onto.C35424,
    onto.C9012,
    onto.C167370,
    onto.C121567,
}

parent_concepts = set()
for c in test_classes:
    parent_concepts |= c.ancestors()
parent_concepts.remove(owl.Thing)
parent_concept_iris = {p.iri for p in parent_concepts}

while element.attrib.get(ABOUT_ATTRIB) != f"{NCIT_PREFIX}#term-group":
    elements = [element]
    element = next(parser)
    while (
        element.tag != CLASS_TAG or ABOUT_ATTRIB not in element.attrib
    ) and element.attrib.get(ABOUT_ATTRIB) != f"{NCIT_PREFIX}#term-group":
        if element.tag == AXIOM_TAG:
            elements.append(element)
        element = next(parser)
    if elements[0].attrib.get(ABOUT_ATTRIB) in parent_concept_iris:
        for e in elements:
            new_root.append(e)
    else:
        for e in elements:
            e.clear()

# get annotations
EOF = False
while not EOF:
    new_root.append(element)
    element = next(parser)
    while True:
        if element.tag == AXIOM_TAG:
            new_root.append(element)
        try:
            element = next(parser)
        except StopIteration:
            EOF = True
            break
        if element.tag == DESCRIPTION_TAG:
            break

ET.ElementTree(new_root).write(outfile_path, pretty_print=True)

pi = XET.ProcessingInstruction(  # TODO get encoding attrib out
    target='xml version="1.0"'
)
pi_string = XET.tostring(pi).decode("ASCII")
with outfile_path.open("r+") as f:
    content = f.read()
    f.seek(0, 0)
    f.write(pi_string.rstrip("\r\n") + "\n" + content)

formatter = xmlformatter.Formatter(indent=2)
formatter.format_file(outfile_path)



"""Contains data models for representing VICC normalized disease records."""

import datetime
from enum import Enum, IntEnum
from types import MappingProxyType
from typing import Literal

from ga4gh.core.models import (
    Coding,
    ConceptMapping,
    MappableConcept,
    Relation,
    code,
)
from pydantic import BaseModel, ConfigDict, StrictBool, StrictStr

from disease import __version__


class MatchType(IntEnum):
    """Define string constraints for use in Match Type attributes."""

    CONCEPT_ID = 100
    LABEL = 80
    ALIAS = 60
    XREF = 60
    ASSOCIATED_WITH = 60
    FUZZY_MATCH = 20
    NO_MATCH = 0


class SourceName(str, Enum):
    """Define string constraints to ensure consistent capitalization."""

    NCIT = "NCIt"
    MONDO = "Mondo"
    DO = "DO"
    ONCOTREE = "OncoTree"
    OMIM = "OMIM"


class SourceIDAfterNamespace(Enum):
    """Define string constraints after namespace."""

    NCIT = "C"
    MONDO = ""
    DO = ""  # noqa: PIE796
    ONCOTREE = ""  # noqa: PIE796
    OMIM = ""  # noqa: PIE796


class NamespacePrefix(Enum):
    """Define string constraints for how concept ID namespace prefixes are
    stored.
    """

    # built-in sources
    NCIT = "ncit"
    MONDO = "mondo"
    DOID = "DOID"
    DO = DOID
    OMIM = "MIM"
    ONCOTREE = "oncotree"
    # external sources
    EFO = "efo"
    GARD = "gard"
    ICD9CM = "icd9.cm"
    ICD10 = "icd10"
    ICD10WHO = ICD10
    ICD10CM = "icd10.cm"
    ICDO = "icdo"
    IMDRF = "imdrf"
    KEGG = "kegg.disease"
    MEDDRA = "meddra"
    MEDGEN = "medgen"
    MESH = "mesh"
    ORPHANET = "orphanet"
    UMLS = "umls"


# Source to URI. Will use  system URI prefix, OBO Foundry persistent URL (PURL), or source homepage
NAMESPACE_TO_SYSTEM_URI: MappingProxyType[NamespacePrefix, str] = MappingProxyType(
    {
        NamespacePrefix.NCIT: "https://ncit.nci.nih.gov/ncitbrowser/ConceptReport.jsp?dictionary=NCI_Thesaurus&code=",
        NamespacePrefix.MONDO: "https://purl.obolibrary.org/obo/",
        NamespacePrefix.DO: "https://disease-ontology.org/?id=",
        NamespacePrefix.DOID: "https://disease-ontology.org/?id=",
        NamespacePrefix.OMIM: "https://omim.org/MIM:",
        NamespacePrefix.ONCOTREE: "https://oncotree.mskcc.org/?version=oncotree_latest_stable&field=CODE&search=",
        NamespacePrefix.EFO: "http://www.ebi.ac.uk/efo/EFO_",
        NamespacePrefix.GARD: "https://rarediseases.info.nih.gov",
        NamespacePrefix.ICD9CM: "https://archive.cdc.gov/www_cdc_gov/nchs/icd/icd9cm.htm",
        NamespacePrefix.ICD10: "https://icd.who.int/browse10/2016/en#/",
        NamespacePrefix.ICD10CM: "https://www.cdc.gov/nchs/icd/icd-10-cm/index.html",
        NamespacePrefix.ICD10WHO: "https://icd.who.int/browse10/2016/en#/",
        NamespacePrefix.ICDO: "https://www.who.int/standards/classifications/other-classifications/international-classification-of-diseases-for-oncology/",
        NamespacePrefix.IMDRF: "https://www.imdrf.org/",
        NamespacePrefix.KEGG: "https://www.genome.jp/kegg/disease/",
        NamespacePrefix.MEDDRA: "https://bioportal.bioontology.org/ontologies/MEDDRA?p=classes&conceptid=",
        NamespacePrefix.MEDGEN: "https://www.ncbi.nlm.nih.gov/medgen/",
        NamespacePrefix.MESH: "https://meshb.nlm.nih.gov/record/ui?ui=",
        NamespacePrefix.ORPHANET: "https://www.orpha.net",
        NamespacePrefix.UMLS: "https://www.nlm.nih.gov/research/umls/index.html",
    }
)


def get_coding_object(concept_id: str) -> Coding:
    """Get coding object for CURIE identifier

    ``system`` will use system prefix URL, OBO Foundry persistent URL (PURL), or
    source homepage, in that order of preference.

    :param concept_id: A lowercase concept identifier represented as a curie
    :raises ValueError: If source of concept ID is not a valid ``NamespacePrefix``
    :return: Coding object for identifier
    """
    source, source_code = concept_id.split(":")

    try:
        source = NamespacePrefix(source)
    except ValueError:
        try:
            source = NamespacePrefix(source.upper())
        except ValueError as e:
            err_msg = f"Namespace prefix not supported: {source}"
            raise ValueError(err_msg) from e

    id_ = concept_id

    if source == NamespacePrefix.MONDO:
        source_code = concept_id.upper()
        id_ = source_code.replace(":", "_")
    elif source == NamespacePrefix.DOID:
        source_code = concept_id

    return Coding(
        id=id_,
        code=code(source_code),
        system=NAMESPACE_TO_SYSTEM_URI[source],
    )


def get_concept_mapping(
    concept_id: str, relation: Relation = Relation.RELATED_MATCH
) -> ConceptMapping:
    """Get concept mapping for CURIE identifier

    :param concept_id: A lowercase concept identifier represented as a curie
    :param relation: SKOS mapping relationship, default is relatedMatch
    :return: Concept mapping for identifier
    """
    return ConceptMapping(
        coding=get_coding_object(concept_id),
        relation=relation,
    )


class SourcePriority(IntEnum):
    """Define priorities for sources in building merged concepts."""

    NCIT = 1
    MONDO = 2
    OMIM = 3
    ONCOTREE = 4
    DO = 5


class Disease(BaseModel):
    """Define disease record."""

    label: StrictStr
    concept_id: StrictStr
    aliases: list[StrictStr] = []
    xrefs: list[StrictStr] = []
    associated_with: list[StrictStr] = []
    pediatric_disease: bool | None = None
    oncologic_disease: bool | None = None

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "label": "Von Hippel-Lindau Syndrome",
                "concept_id": "ncit:C3105",
                "aliases": [
                    "Von Hippel-Lindau Syndrome (VHL)",
                    "Von Hippel-Lindau Disease",
                    "Cerebroretinal Angiomatosis",
                    "von Hippel-Lindau syndrome",
                    "VHL syndrome",
                ],
                "xrefs": [],
                "associated_with": ["umls:C0019562"],
                "pediatric_disease": None,
                "oncologic_disease": None,
            }
        }
    )


class RecordType(str, Enum):
    """Record item types."""

    IDENTITY = "identity"
    MERGER = "merger"


class RefType(str, Enum):
    """Reference item types."""

    # Must be in descending MatchType order.
    LABEL = "label"
    ALIASES = "alias"
    XREFS = "xref"
    ASSOCIATED_WITH = "associated_with"


class DataLicenseAttributes(BaseModel):
    """Define constraints for data license attributes."""

    non_commercial: StrictBool
    share_alike: StrictBool
    attribution: StrictBool


class SourceMeta(BaseModel):
    """Metadata for a given source to return in response object."""

    data_license: StrictStr
    data_license_url: StrictStr
    version: StrictStr
    data_url: StrictStr | None = None
    rdp_url: StrictStr | None = None
    data_license_attributes: DataLicenseAttributes

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "data_license": "CC BY 4.0",
                "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                "version": "21.01d",
                "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/21.01d_Release/",
                "rdp_url": "http://reusabledata.org/ncit.html",
                "data_license_attributes": {
                    "non_commercial": False,
                    "attribution": True,
                    "share_alike": False,
                },
            }
        }
    )


class SourceSearchMatches(BaseModel):
    """Container for matching information from an individual source."""

    match_type: MatchType
    records: list[Disease] = []
    source_meta_: SourceMeta

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "match_type": 80,
                "records": [
                    {
                        "label": "non-small cell lung carcinoma",
                        "concept_id": "mondo:0005233",
                        "aliases": [
                            "non-small cell lung carcinoma (disease)",
                            "non-small cell carcinoma of lung",
                            "non-small cell lung cancer",
                            "non-small cell cancer of lung",
                            "non-small cell carcinoma of the lung",
                            "non-small cell cancer of the lung",
                            "NSCLC",
                            "NSCLC - non-small cell lung cancer",
                        ],
                        "xrefs": ["DOID:3908", "ncit:C2926", "oncotree:NSCLC"],
                        "associated_with": [
                            "mesh:D002289",
                            "efo:0003060",
                            "umls:C0007131",
                        ],
                    }
                ],
                "source_meta_": {
                    "data_license": "CC BY 4.0",
                    "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                    "version": "2023-07-03",
                    "data_url": "https://mondo.monarchinitiative.org/pages/download/",
                    "rdp_url": "http://reusabledata.org/monarch.html",
                    "data_license_attributes": {
                        "non_commercial": False,
                        "attribution": True,
                        "share_alike": False,
                    },
                },
            }
        }
    )


class ServiceMeta(BaseModel):
    """Metadata regarding the disease-normalization service."""

    name: Literal["disease-normalizer"] = "disease-normalizer"
    version: StrictStr
    response_datetime: datetime.datetime
    url: Literal["https://github.com/cancervariants/disease-normalization"] = (
        "https://github.com/cancervariants/disease-normalization"
    )

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "name": "disease-normalizer",
                "version": __version__,
                "response_datetime": "2021-04-05T16:44:15.367831",
                "url": "https://github.com/cancervariants/disease-normalization",
            }
        }
    )


class NormalizationService(BaseModel):
    """Response containing one or more merged records and source data."""

    query: StrictStr
    warnings: dict | None = None
    match_type: MatchType
    disease: MappableConcept | None = None
    source_meta_: dict[SourceName, SourceMeta] | None = None
    service_meta_: ServiceMeta

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "query": "childhood leukemia",
                "warnings": None,
                "match_type": 80,
                "disease": {
                    "id": "normalize.disease.ncit:C4989",
                    "primaryCoding": {
                        "id": "ncit:C4989",
                        "code": "C4989",
                        "system": "https://ncit.nci.nih.gov/ncitbrowser/ConceptReport.jsp?dictionary=NCI_Thesaurus&code=",
                    },
                    "conceptType": "Disease",
                    "name": "Childhood Leukemia",
                    "mappings": [
                        {
                            "coding": {
                                "id": "MONDO_0004355",
                                "code": "MONDO:0004355",
                                "system": "https://purl.obolibrary.org/obo/",
                            },
                            "relation": "exactMatch",
                        },
                        {
                            "coding": {
                                "id": "DOID:7757",
                                "code": "DOID:7757",
                                "system": "https://disease-ontology.org/?id=",
                            },
                            "relation": "exactMatch",
                        },
                        {
                            "coding": {
                                "id": "umls:C1332977",
                                "code": "C1332977",
                                "system": "https://www.nlm.nih.gov/research/umls/index.html",
                            },
                            "relation": "relatedMatch",
                        },
                    ],
                    "extensions": [
                        {
                            "name": "aliases",
                            "value": [
                                "childhood leukemia (disease)",
                                "leukemia",
                                "pediatric leukemia (disease)",
                                "Leukemia",
                                "leukemia (disease) of childhood",
                            ],
                        },
                        {
                            "name": "pediatric_disease",
                            "value": True,
                        },
                    ],
                },
                "source_meta_": {
                    "NCIt": {
                        "data_license": "CC BY 4.0",
                        "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                        "version": "21.01d",
                        "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/21.01d_Release/",
                        "rdp_url": "http://reusabledata.org/ncit.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False,
                        },
                    },
                    "Mondo": {
                        "data_license": "CC BY 4.0",
                        "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
                        "version": "20210129",
                        "data_url": "https://mondo.monarchinitiative.org/pages/download/",
                        "rdp_url": "http://reusabledata.org/monarch.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False,
                        },
                    },
                    "DO": {
                        "data_license": "CC0 1.0",
                        "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/legalcode",
                        "version": "20210305",
                        "data_url": "http://www.obofoundry.org/ontology/doid.html",
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": False,
                            "share_alike": False,
                        },
                    },
                },
                "service_meta_": {
                    "name": "disease-normalizer",
                    "version": __version__,
                    "response_datetime": "2021-04-05T16:44:15.367831",
                    "url": "https://github.com/cancervariants/disease-normalization",
                },
            }
        }
    )


class SearchService(BaseModel):
    """Core response schema containing matches for each source"""

    query: StrictStr
    warnings: dict | None = None
    source_matches: dict[SourceName, SourceSearchMatches]
    service_meta_: ServiceMeta

    model_config = ConfigDict(
        json_schema_extra={
            "query": "nsclc",
            "source_matches": {
                "OMIM": {
                    "records": [],
                    "source_meta_": {
                        "data_license": "custom",
                        "data_license_url": "https://omim.org/help/agreement",
                        "version": "20231030",
                        "data_url": "https://www.omim.org/downloads",
                        "rdp_url": "http://reusabledata.org/omim.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": True,
                        },
                    },
                },
                "DO": {
                    "records": [
                        {
                            "label": "lung non-small cell carcinoma",
                            "concept_id": "DOID:3908",
                            "aliases": [
                                "Non-small cell lung cancer",
                                "NSCLC",
                                "non-small cell lung carcinoma",
                            ],
                            "xrefs": ["ncit:C2926"],
                            "associated_with": [
                                "mesh:D002289",
                                "kegg.disease:05223",
                                "efo:0003060",
                                "umls:C0007131",
                            ],
                            "match_type": 60,
                        }
                    ],
                    "source_meta_": {
                        "data_license": "CC0 1.0",
                        "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/legalcode",
                        "version": "2023-07-24",
                        "data_url": "http://www.obofoundry.org/ontology/doid.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": False,
                            "share_alike": False,
                        },
                    },
                },
            },
            "service_meta_": {
                "name": "disease-normalizer",
                "version": "0.4.0.dev1",
                "response_datetime": "2023-10-31T10:53:30.890262",
                "url": "https://github.com/cancervariants/disease-normalization",
            },
        }
    )


class ServiceEnvironment(str, Enum):
    """Define current runtime environment."""

    DEV = "dev"  # local dev
    TEST = "test"  # local/CI testing
    PROD = "prod"  # main production env on cloud
    STAGING = "staging"  # staging env on cloud


LAB_EMAIL = "Alex.Wagner@nationwidechildrens.org"
LAB_WEBPAGE_URL = "https://www.nationwidechildrens.org/specialties/institute-for-genomic-medicine/research-labs/wagner-lab"
DOCS_URL = "https://disease-normalizer.readthedocs.io/"
APP_DESCRIPTION = "Resolve ambiguous references and descriptions of human diseases to consistently-structured, normalized terms"


class ServiceType(BaseModel):
    """Define service_info response for type field"""

    group: Literal["org.genomicmedlab"] = "org.genomicmedlab"
    artifact: Literal["Disease Normalizer API"] = "Disease Normalizer API"
    version: Literal[__version__] = __version__


class ServiceOrganization(BaseModel):
    """Define service_info response for organization field"""

    name: Literal["Genomic Medicine Lab at Nationwide Children's Hospital"] = (
        "Genomic Medicine Lab at Nationwide Children's Hospital"
    )
    url: Literal[LAB_WEBPAGE_URL] = LAB_WEBPAGE_URL


class ServiceInfo(BaseModel):
    """Define response structure for GA4GH /service_info endpoint."""

    id: Literal["org.genomicmedlab.disease_normalizer"] = (
        "org.genomicmedlab.disease_normalizer"
    )
    name: Literal["disease_normalizer"] = "disease_normalizer"
    type: ServiceType
    description: Literal[APP_DESCRIPTION] = APP_DESCRIPTION
    organization: ServiceOrganization
    contactUrl: Literal[LAB_EMAIL] = LAB_EMAIL  # noqa: N815
    documentationUrl: Literal[DOCS_URL] = DOCS_URL  # noqa: N815
    createdAt: Literal["2021-01-01T00:00:00+00:00"] = "2021-01-01T00:00:00+00:00"  # noqa: N815
    updatedAt: Literal["2021-01-01T00:00:00+00:00"] = "2021-01-01T00:00:00+00:00"  # noqa: N815
    environment: ServiceEnvironment
    version: Literal[__version__] = __version__



"""Create concept groups and merged records."""

import logging
from collections.abc import Collection
from timeit import default_timer as timer

from tqdm import tqdm

from disease.database.database import AbstractDatabase
from disease.schemas import SourcePriority

_logger = logging.getLogger(__name__)


class Merge:
    """Manage construction of record mergers for normalization."""

    def __init__(self, database: AbstractDatabase, silent: bool = True) -> None:
        """Initialize Merge instance.

        :param Database database: db instance to use for record retrieval and creation.
        :param silent: if ``True``, suppress console output
        """
        self._database = database
        self._silent = silent
        self._groups = []  # list of tuples: (mondo_concept_id, set_of_ids)

    def create_merged_concepts(self, record_ids: Collection[str]) -> None:
        """Create concept groups, generate merged concept records, and update database.

        Our normalization protocols only generate record ID sets that include Mondo
        terms, meaning only Mondo IDs should be submitted to this method.

        Rather than a DFS of all record xrefs as with the other normalizers, here we just
        work from the xrefs provided by the Mondo record.

        :param record_ids: concept identifiers from which groups should be generated.
        """
        # build groups
        _logger.info("Generating record ID sets from %s records", len(record_ids))
        start = timer()
        for concept_id in tqdm(record_ids, ncols=80, disable=self._silent):
            try:
                record = self._database.get_record_by_id(concept_id)
            except AttributeError:
                _logger.exception(
                    "`create_merged_concepts` received invalid concept ID: %s",
                    concept_id,
                )
                continue
            if not record:
                _logger.error("generate_merged_concepts couldn't find %s", concept_id)
                continue
            xrefs = record.get("xrefs", None)
            group = {*xrefs, concept_id} if xrefs else {concept_id}
            self._groups.append((concept_id, group))
        end = timer()
        self._database.complete_write_transaction()
        _logger.debug("Built record ID sets in %s seconds", end - start)

        # build merged concepts
        _logger.info("Creating merged records and updating database...")
        start = timer()
        for record_id, group in tqdm(self._groups, ncols=80, disable=self._silent):
            try:
                merged_record, merged_ids = self._generate_merged_record(group)
            except AttributeError:
                _logger.exception(
                    "`create_merged_concepts` received invalid group: %s for concept %s",
                    group,
                    record_id,
                )
                continue
            self._database.add_merged_record(merged_record)
            merge_ref = merged_record["concept_id"]

            for concept_id in merged_ids:
                self._database.update_merge_ref(concept_id, merge_ref)
        self._database.complete_write_transaction()
        end = timer()
        _logger.info("merged concept generation successful.")
        _logger.debug("Generated and added concepts in %s seconds", end - start)

    def _generate_merged_record(self, record_id_set: set[str]) -> tuple[dict, list]:
        """Generate merged record from provided concept ID group.

        Misc notes
        * Where attributes are sets, they should be merged, and where they are
        scalars, assign from the highest-priority source where that attribute
        is non-null. Value priority is NCIt > Mondo > OMIM > OncoTree> DO.
        * The exception is with xrefs, which (for all groups of n > 1) is constructed by
        taking the `record_id_set` value, dropping all records which (for whatever reason)
        weren't able to be found in the DB, and dropping the ID that will be used as the
        primary coding for the record

        :param Set record_id_set: group of concept IDs
        :return: completed merged drug object to be stored in DB, as well as
            a list of the IDs ultimately included in said record
        """
        records = []
        final_ids = []
        for record_id in record_id_set:
            record = self._database.get_record_by_id(record_id)
            if record:
                records.append(record)
                final_ids.append(record["concept_id"])
            else:
                _logger.error(
                    "generate_merged_record could not retrieve record for %s in %s",
                    record_id,
                    record_id_set,
                )

        def record_order(record: dict) -> tuple:
            """Provide priority values of concepts for sort function."""
            src = record["src_name"].upper()
            source_rank = SourcePriority[src].value
            return source_rank, record["concept_id"]

        records.sort(key=record_order)

        merged_properties = {
            "concept_id": records[0]["concept_id"],
            "aliases": set(),
            "associated_with": set(),
        }
        if len(records) > 1:
            merged_properties["xrefs"] = list({r["concept_id"] for r in records[1:]})

        set_fields = ["aliases", "associated_with"]
        scalar_fields = ["label", "pediatric_disease", "oncologic_disease"]
        for record in records:
            for field in set_fields:
                if field in record:
                    merged_properties[field] |= set(record[field])
            for field in scalar_fields:
                if field not in merged_properties and field in record:
                    merged_properties[field] = record[field]

        for field in set_fields:
            field_value = merged_properties[field]
            if field_value:
                merged_properties[field] = list(field_value)
            else:
                del merged_properties[field]

        return merged_properties, final_ids



"""Construct test data for DO source."""

import contextlib
import os
import subprocess
from http import HTTPStatus
from pathlib import Path

import requests

from disease.database import create_db
from disease.etl import DO

do = DO(create_db())
do._extract_data()

infile = do._data_file.absolute()

scripts_dir = Path(__file__).resolve().parent
test_data_dir = scripts_dir.parent / "data" / "do"

robot_file = scripts_dir / "robot"
if not robot_file.exists():
    response = requests.get(
        "https://raw.githubusercontent.com/ontodev/robot/master/bin/robot", timeout=30
    )
    if response.status_code != HTTPStatus.OK:
        msg = "Couldn't acquire robot script"
        raise requests.HTTPError(msg)
    with robot_file.open("wb") as f:
        f.write(response.content)
    with contextlib.suppress(PermissionError):
        robot_file.chmod(0o755)
if not os.access(robot_file, os.X_OK):
    msg = "robot file isn't executable by the user -- see 'getting started': http://robot.obolibrary.org/"
    raise PermissionError(msg)

robot_jar = scripts_dir / "robot.jar"
if not robot_jar.exists():
    response = requests.get(
        "https://api.github.com/repos/ontodev/robot/releases/latest", timeout=30
    )
    if response.status_code != HTTPStatus.OK:
        msg = "Couldn't get ROBOT release info from GitHub"
        raise requests.HTTPError(msg)
    json = response.json()
    assert json["assets"][0]["name"] == "robot.jar"
    jar_url = json["assets"][0]["url"]
    jar_response = requests.get(jar_url, timeout=30)
    if jar_response.status_code != HTTPStatus.OK:
        msg = "Couldn't download ROBOT JAR from GitHub"
        raise requests.HTTPError(msg)
    with robot_jar.open("wb") as f:
        f.write(jar_response.content)

terms_file = scripts_dir / "do_terms.txt"
if not terms_file.exists():
    msg = "Could not find do_terms.txt"
    raise FileNotFoundError(msg)

outfile = test_data_dir / do._data_file.name
outfile.parent.mkdir(exist_ok=True)
cmd_str = f"{robot_file} extract --method star --input {infile} --term-file {terms_file} --output {outfile}"

subprocess.run(cmd_str, shell=True)  # noqa: S602



"""Test utils module"""

import pytest
from deepdiff import DeepDiff

from disease.database.database import AbstractDatabase
from disease.schemas import RecordType, SourceName
from disease.utils import get_term_mappings


@pytest.fixture(scope="module")
def test_db(null_database_class):
    return null_database_class(
        get_all_records={
            RecordType.IDENTITY: [
                {
                    "concept_id": "ncit:C185069",
                    "label": "Metastatic Hilar Cholangiocarcinoma",
                    "src_name": "NCIt",
                    "item_type": "identity",
                },
                {
                    "concept_id": "DOID:10348",
                    "label": "blepharophimosis",
                    "associated_with": [
                        "mesh:D016569",
                        "icd9.cm:374.46",
                        "icd10.cm:H02.52",
                        "umls:C0005744",
                        "gard:5932",
                    ],
                    "src_name": "DO",
                    "merge_ref": "mondo:0001008",
                    "item_type": "identity",
                },
                {
                    "concept_id": "ncit:C6610",
                    "label": "Stage IIA Osteosarcoma AJCC v7",
                    "aliases": [
                        "Stage IIA Osteosarcoma",
                        "Stage IIA Osteogenic Sarcoma",
                    ],
                    "associated_with": ["umls:C1336168"],
                    "src_name": "NCIt",
                    "item_type": "identity",
                },
                {
                    "concept_id": "mondo:0018648",
                    "label": "Keratocystic odontogenic tumor",
                    "aliases": [
                        "odontogenic keratocystoma",
                        "odontogenic Keratocyst",
                        "KTOC",
                    ],
                    "associated_with": [
                        "umls:C1708604",
                        "medgen:313330",
                        "orphanet:447777",
                    ],
                    "xrefs": ["ncit:C54302"],
                    "src_name": "Mondo",
                    "merge_ref": "mondo:0018648",
                    "oncologic_disease": True,
                    "item_type": "identity",
                },
            ],
            RecordType.MERGER: [
                {
                    "concept_id": "mondo:0008213",
                    "label": "pectus excavatum",
                    "aliases": [
                        "funnel chest",
                        "pectus excavatum",
                        "FUNNEL CHEST",
                        "pectus excavatum (disease)",
                    ],
                    "associated_with": [
                        "umls:C2051831",
                        "mesh:D005660",
                        "icd10.cm:Q67.6",
                        "medgen:781174",
                    ],
                    "xrefs": ["MIM:169300"],
                    "item_type": "merger",
                },
                {
                    "concept_id": "ncit:C40144",
                    "label": "Endometrial Mucinous Adenocarcinoma",
                    "aliases": [
                        "endometrium mucinous adenocarcinoma",
                        "Uterine Corpus Mucinous Adenocarcinoma",
                        "uterine mucinous carcinoma",
                        "uterine corpus mucinous adenocarcinoma",
                        "endometrial mucinous adenocarcinoma",
                        "uterine Corpus mucinous adenocarcinoma",
                    ],
                    "associated_with": [
                        "medgen:276939",
                        "umls:C1519859",
                        "efo:1000236",
                        "umls:C0854923",
                    ],
                    "xrefs": ["mondo:0002747", "oncotree:UMC", "DOID:3707"],
                    "oncologic_disease": True,
                    "item_type": "merger",
                },
                {
                    "concept_id": "ncit:C5063",
                    "label": "Endobronchial Lipoma",
                    "aliases": ["endobronchial lipoma"],
                    "associated_with": ["medgen:208874", "umls:C0852937"],
                    "xrefs": ["DOID:10183", "mondo:0000961"],
                    "oncologic_disease": True,
                    "item_type": "merger",
                },
                {
                    "concept_id": "oncotree:CAEXPA",
                    "label": "Carcinoma ex Pleomorphic Adenoma",
                    "src_name": "OncoTree",
                    "oncologic_disease": True,
                    "item_type": "identity",
                },
            ],
        }
    )


def test_get_term_mappings_ncit(test_db: AbstractDatabase):
    results = list(get_term_mappings(test_db, SourceName.NCIT))

    ncit_mappings_fixture = [
        {
            "concept_id": "ncit:C185069",
            "label": "Metastatic Hilar Cholangiocarcinoma",
            "xrefs": [],
            "aliases": [],
        },
        {
            "concept_id": "ncit:C6610",
            "label": "Stage IIA Osteosarcoma AJCC v7",
            "xrefs": ["umls:C1336168"],
            "aliases": [
                "Stage IIA Osteosarcoma",
                "Stage IIA Osteogenic Sarcoma",
            ],
        },
    ]
    diff = DeepDiff(ncit_mappings_fixture, results, ignore_order=True)
    assert diff == {}


def test_get_term_mappings_merger(test_db: AbstractDatabase):
    results = list(get_term_mappings(test_db, scope=RecordType.MERGER))

    fixture = [
        {
            "concept_id": "mondo:0008213",
            "label": "pectus excavatum",
            "aliases": [
                "funnel chest",
                "pectus excavatum",
                "FUNNEL CHEST",
                "pectus excavatum (disease)",
            ],
            "xrefs": [
                "umls:C2051831",
                "mesh:D005660",
                "icd10.cm:Q67.6",
                "medgen:781174",
                "MIM:169300",
            ],
        },
        {
            "concept_id": "ncit:C40144",
            "label": "Endometrial Mucinous Adenocarcinoma",
            "aliases": [
                "endometrium mucinous adenocarcinoma",
                "Uterine Corpus Mucinous Adenocarcinoma",
                "uterine mucinous carcinoma",
                "uterine corpus mucinous adenocarcinoma",
                "endometrial mucinous adenocarcinoma",
                "uterine Corpus mucinous adenocarcinoma",
            ],
            "xrefs": [
                "medgen:276939",
                "umls:C1519859",
                "efo:1000236",
                "umls:C0854923",
                "mondo:0002747",
                "oncotree:UMC",
                "DOID:3707",
            ],
        },
        {
            "concept_id": "ncit:C5063",
            "label": "Endobronchial Lipoma",
            "aliases": ["endobronchial lipoma"],
            "xrefs": ["medgen:208874", "umls:C0852937", "DOID:10183", "mondo:0000961"],
        },
        {
            "concept_id": "oncotree:CAEXPA",
            "label": "Carcinoma ex Pleomorphic Adenoma",
            "aliases": [],
            "xrefs": [],
        },
    ]
    diff = DeepDiff(fixture, results, ignore_order=True)
    assert diff == {}


def test_get_term_mappings_cancer(test_db: AbstractDatabase):
    results = list(
        get_term_mappings(test_db, scope=RecordType.IDENTITY, cancer_only=True)
    )
    assert len(results) == 1



"""Provide functions to perform Disease Normalizer updates."""

import logging
from timeit import default_timer as timer

import click

from disease.database.database import (
    AbstractDatabase,
    DatabaseReadException,
    DatabaseWriteException,
)
from disease.schemas import SourceName

_logger = logging.getLogger(__name__)


def _emit_info_msg(msg: str, silent: bool) -> None:
    """Handle info message display.

    :param msg: message to log/print
    :param silent: if True, don't print to console
    """
    if not silent:
        click.echo(msg)
    _logger.info(msg)


def delete_source(
    source: SourceName, db: AbstractDatabase, silent: bool = True
) -> float:
    """Delete all data for an individual source.

    :param source: name of source to delete data for
    :param db: database instance
    :param silent: if True, suppress console output
    :return: time spent deleting source
    """
    _emit_info_msg(f"Deleting {source.value}...", silent)
    start_delete = timer()
    db.delete_source(source)
    end_delete = timer()
    delete_time = end_delete - start_delete
    _emit_info_msg(f"Deleted {source.value} in {delete_time:.5f} seconds.", silent)
    return delete_time


_etl_dependency_help = "Are ETL dependencies installed? See the Installation page in the documentation for more info."


def load_source(
    source: SourceName, db: AbstractDatabase, use_existing: bool, silent: bool = True
) -> tuple[float, set[str]]:
    """Load data for an individual source.

    :param source: name of source to load data for
    :param db: database instance
    :param use_existing: if True, use latest available version of local data
    :param silent: if True, suppress console output
    :return: time spent loading data, and set of processed IDs from that source
    """
    _emit_info_msg(f"Loading {source.value}...", silent)
    start_load = timer()

    # used to get source class name from string
    try:
        from disease.etl import (  # noqa: PLC0415
            DO,
            OMIM,
            DiseaseNormalizerEtlError,
            Mondo,
            NCIt,
            OncoTree,
        )
    except ModuleNotFoundError as e:
        click.echo(
            f"Encountered ModuleNotFoundError attempting to import {e.name}. {_etl_dependency_help}"
        )
        click.get_current_context().exit()
    sources_table = {
        SourceName.DO: DO,
        SourceName.MONDO: Mondo,
        SourceName.NCIT: NCIt,
        SourceName.OMIM: OMIM,
        SourceName.ONCOTREE: OncoTree,
    }

    source_class = sources_table[source](database=db, silent=silent)
    try:
        processed_ids = source_class.perform_etl(use_existing)
    except DiseaseNormalizerEtlError as e:
        msg = f"Encountered error while loading {source}: {e}."
        _logger.exception(msg)
        if not silent:
            click.echo(msg)
        click.get_current_context().exit(1)
    end_load = timer()
    load_time = end_load - start_load
    _emit_info_msg(
        f"Loaded {len(processed_ids)} records from {source.value} in {load_time:.5f} seconds.",
        silent,
    )

    # don't forward any concept IDs that we don't want to use for group generation
    if not source_class.store_ids:
        processed_ids = []

    return (load_time, set(processed_ids))


def update_source(
    source: SourceName, db: AbstractDatabase, use_existing: bool, silent: bool = True
) -> set[str]:
    """Refresh data for an individual disease data source.

    For example, to completely refresh Mondo data:

    >>> from disease.schemas import SourceName
    >>> from disease.database import create_db
    >>> from disease.etl.update import update_source
    >>> db = create_db()
    >>> processed_ids = update_source(SourceName.MONDO, db)

    :param source: name of source to update
    :param db: database instance
    :param use_existing: if True, use latest available local data
    :param silent: if True, suppress console output
    :return: IDs for records created from source
    """
    delete_time = delete_source(source, db, silent)
    load_time, processed_ids = load_source(source, db, use_existing, silent)
    _emit_info_msg(
        f"Total time for {source.value}: {(delete_time + load_time):.5f} seconds.",
        silent,
    )
    return processed_ids


def update_all_sources(
    db: AbstractDatabase, use_existing: bool, silent: bool = True
) -> set[str]:
    """Refresh data for all disease record sources.

    :param db: database instance
    :param use_existing: if True, use latest available local data for all sources
    :param silent: if True, suppress console output
    :return: IDs processed from all sources
    """
    processed_ids: list[str] = []
    for source in SourceName:
        source_ids = update_source(source, db, use_existing, silent)
        processed_ids += list(source_ids)
    return set(processed_ids)


def delete_normalized(database: AbstractDatabase, silent: bool = True) -> None:
    """Delete normalized concepts.

    :param database: DB instance
    :param silent: if True, suppress console output
    """
    _emit_info_msg("\nDeleting normalized records...", silent)
    start_delete = timer()
    try:
        database.delete_normalized_concepts()
    except (DatabaseReadException, DatabaseWriteException):
        msg = "Encountered exception during normalized data deletion"
        _logger.exception(msg)
        click.echo(msg)
        raise
    end_delete = timer()
    delete_time = end_delete - start_delete
    _emit_info_msg(f"Deleted normalized records in {delete_time:.5f} seconds.", silent)


def update_normalized(
    db: AbstractDatabase, processed_ids: set[str] | None, silent: bool = True
) -> None:
    """Delete existing and update merged normalized records.

    :param db: database instance
    :param processed_ids: IDs to form normalized records from. Provide if available to
        cut down on some potentially slow database calls. If unavailable, this method
        will fetch all known IDs directly.
    :param silent: if True, suppress console output
    """
    start = timer()
    delete_normalized(db, silent)
    if not processed_ids:
        processed_ids = db.get_all_concept_ids()

    try:
        from disease.etl.merge import Merge  # noqa: PLC0415
    except ModuleNotFoundError as e:
        msg = f"Encountered ModuleNotFoundError attempting to import {e.name}. {_etl_dependency_help}"
        if not silent:
            click.echo(msg)
        _logger.exception(msg)
        click.get_current_context().exit()

    merge = Merge(database=db)
    if not silent:
        click.echo("Constructing normalized records...")
    merge.create_merged_concepts(processed_ids)
    end = timer()
    _emit_info_msg(
        f"Merged concept generation completed in {(end - start):.5f} seconds",
        silent,
    )


def update_all_and_normalize(
    db: AbstractDatabase, use_existing: bool, silent: bool = True
) -> None:
    """Update all sources as well as normalized records.

    For example, to completely refresh all Disease Normalizer data:

    >>> from disease.database import create_db
    >>> from disease.etl.update import update_all_and_normalize
    >>> db = create_db()
    >>> update_all_and_normalize(db, False)

    :param db: database instance
    :param use_existing: if True, use latest local copy of data
    :param silent: if True, suppress console output
    """
    processed_ids = update_all_sources(db, use_existing, silent)
    update_normalized(db, processed_ids, silent)



"""Build OncoTree test data.

It's actually just the real data, because it's not too big.
"""

import shutil
from pathlib import Path

from disease.database import create_db
from disease.etl import OncoTree

oncotree = OncoTree(create_db())
oncotree._extract_data()
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "oncotree"
outfile_path = TEST_DATA_DIR / oncotree._data_file.name
outfile_path.parent.mkdir(exist_ok=True)

shutil.copyfile(oncotree._data_file, outfile_path)
